"""
obiKEGG2 configuration

mostly just caching settings

"""
from __future__ import absolute_import

import os

try:
    import ConfigParser as configparser
except ImportError:
    import configparser

from six import StringIO
from ..utils import serverfiles
kegg_dir = serverfiles.localpath("KEGG2")

default = """
[cache]
# path = %(home)s/.obiKEGG/
path = %(kegg_dir)s/
store = sqlite3
invalidate = weekly

[service]
transport = urllib2
# transport = requests

"""

# Orange kegg files dir

env = dict(os.environ)
env["kegg_dir"] = kegg_dir

parser = configparser.ConfigParser(env)


parser.readfp(StringIO(default), "default")

# TODO: global settings rc file
parser.read([os.path.expanduser("~/.obiKEGG/rc.cfg")])

params = {}

_ALL_PARAMS = [
    "cache.path",
    "cache.store",
    "cache.invalidate",
    "service.transport"
]

for p in _ALL_PARAMS:
    section, option = p.split(".")
    params[p] = parser.get(section, option)



import sqlite3
import bz2
import pickle


from collections import defaultdict
from server_update import *
from server_update.tests.test_GeneInfo import GeneInfo
from orangecontrib.bioinformatics.ncbi.gene import (
    DOMAIN, FILENAME, MATCHER_FILENAME, MATCHER_TITLE, MATCHER_TAGS,
    MAP_GENE_ID, MAP_SOURCES, MAP_SYMBOL, MAP_SYNONYMS, MAP_LOCUS, MAP_NOMENCLATURE, MAP_TAX_ID)

from orangecontrib.bioinformatics.ncbi.gene.utils import parse_sources, parse_synonyms, GeneInfoDB
from orangecontrib.bioinformatics.ncbi.taxonomy import common_taxids, common_taxid_to_name
from orangecontrib.bioinformatics.utils import serverfiles


serverfiles.localpath_download(DOMAIN, FILENAME)

# for indexing results from gene info database
tax_id, gene_id, symbol, synonyms, sources, locus_tag, nomenclature = range(7)

domain_path = sf_local.localpath(DOMAIN)
temp_path = os.path.join(domain_path, sf_temp)

db_path = os.path.join(domain_path, FILENAME)


create_folder(temp_path)
create_folder(domain_path)


def parse_gene_record(parent_tax, mapper, gene_record):
    gene = {MAP_TAX_ID:        parent_tax,
            MAP_GENE_ID:       gene_record[gene_id],
            MAP_SYMBOL:        gene_record[symbol],
            MAP_SYNONYMS:      parse_synonyms(gene_record[synonyms]),
            MAP_SOURCES:       parse_sources(gene_record[sources]),
            MAP_LOCUS:         gene_record[locus_tag],
            MAP_NOMENCLATURE:  gene_record[nomenclature]
            }

    # construct gene mapper
    mapper[MAP_SYMBOL][gene[MAP_SYMBOL]].append(gene)
    mapper[MAP_LOCUS][gene[MAP_LOCUS]].append(gene)
    mapper[MAP_GENE_ID][gene[MAP_GENE_ID]].append(gene)
    mapper[MAP_NOMENCLATURE][gene[MAP_NOMENCLATURE]].append(gene)

    for gene_synonym in gene[MAP_SYNONYMS]:
        mapper[MAP_SYNONYMS][gene_synonym].append(gene)

    for source_id in gene[MAP_SOURCES].values():
        mapper[MAP_SOURCES][source_id].append(gene)


print("Creating gene name mapper ...")

con = sqlite3.connect(db_path, timeout=15)
cursor = con.cursor()

for taxonomy_id in common_taxids():
    g_db = GeneInfoDB()
    gene_mapper = {MAP_GENE_ID: defaultdict(list),
                   MAP_SOURCES: defaultdict(list),
                   MAP_SYMBOL: defaultdict(list),
                   MAP_SYNONYMS: defaultdict(list),
                   MAP_LOCUS: defaultdict(list),
                   MAP_NOMENCLATURE: defaultdict(list)}

    for record in g_db.select_gene_matcher_data(taxonomy_id):
        parse_gene_record(taxonomy_id, gene_mapper, record)

    with open(os.path.join(domain_path, MATCHER_FILENAME.format(taxonomy_id)), 'wb') as file:
        pickle.dump(gene_mapper, file, protocol=pickle.HIGHEST_PROTOCOL)

    uncompressed_size = os.stat(os.path.join(domain_path, MATCHER_FILENAME.format(taxonomy_id))).st_size

    with bz2.BZ2File(os.path.join(temp_path, MATCHER_FILENAME.format(taxonomy_id)), mode='w', compresslevel=9) as f:
        shutil.copyfileobj(open(os.path.join(domain_path, MATCHER_FILENAME.format(taxonomy_id)), "rb"), f)

    create_info_file(os.path.join(temp_path, MATCHER_FILENAME.format(taxonomy_id)),
                     domain=DOMAIN,
                     filename=MATCHER_FILENAME.format(taxonomy_id),
                     source=SOURCE_SERVER,
                     title=MATCHER_TITLE + ' for ' + common_taxid_to_name(taxonomy_id),
                     tags=MATCHER_TAGS + [taxonomy_id],
                     uncompressed=uncompressed_size,
                     compression='bz2')


con.close()

helper = SyncHelper(DOMAIN, GeneInfo)

# sync files with remote server
helper.run_tests()
helper.sync_files()

helper.remove_update_folder()



import unittest
from os import makedirs, path

from Orange.data import Table

from orangecontrib.bioinformatics.geo.utils import (
    gds_is_cached, gds_download_url, gds_ensure_downloaded, gds_download,
    spots_mean, spots_median, spots_min, spots_max, unknown, parse_attribute_line, DOMAIN
)
from orangecontrib.bioinformatics.geo.dataset import GDSInfo, GDS
from orangecontrib.bioinformatics.utils import serverfiles


class TestGEO(unittest.TestCase):

    test_sample = 'GDS1001'
    test_organism = '10090'

    def test_gds_info(self):
        gds_info = GDSInfo()
        self.assertIsNotNone(gds_info)
        self.assertGreater(len(gds_info.keys()), 0)
        self.assertGreater(len(gds_info.items()), 0)
        self.assertGreater(len(gds_info.values()), 0)

        self.assertIsNotNone(gds_info[self.test_sample])
        self.assertEqual(gds_info[self.test_sample]['gene_count'], 9561)
        self.assertEqual(len(gds_info[self.test_sample]['samples']), 4)
        self.assertEqual(len(gds_info[self.test_sample]['subsets']), 2)

    def test_gds_utils(self):
        self.assertEqual(spots_mean([2, 2, 8, 8]), 5)
        self.assertEqual(spots_mean([8, unknown, 4, unknown]), 6)
        self.assertIsInstance(spots_mean([unknown, unknown]), float)

        self.assertEqual(spots_median([2, 2, 4, 8, 8]), 4)
        self.assertEqual(spots_median([8, unknown, 4, unknown]), 6)
        self.assertIsInstance(spots_median([unknown, unknown]), float)

        self.assertEqual(spots_min([2, 2, 8, 8]), 2)
        self.assertEqual(spots_min([8, unknown, 4, unknown]), 4)
        self.assertIsInstance(spots_min([unknown, unknown]), float)

        self.assertEqual(spots_max([2, 2, 8, 8]), 8)
        self.assertEqual(spots_max([8, unknown, 4, unknown]), 8)
        self.assertIsInstance(spots_max([unknown, unknown]), float)

        self.assertEqual(parse_attribute_line('!dataset_title = test_title'), ('title', 'test_title'))
        self.assertEqual(parse_attribute_line('!dataset_description = test_desc'),  ('description', 'test_desc'))
        self.assertEqual(parse_attribute_line('!dataset_gds_type = test_type'), ('gds_type', 'test_type'))

        # header like line
        self.assertRaises(AttributeError, parse_attribute_line, 'ID  SAMPLE1 SAMPLE2 SAMPLE3')

    def test_gds_data(self):
        # test url
        self.assertIsNotNone(gds_download_url(self.test_sample))

        # file not in cache
        self.assertFalse(gds_is_cached(self.test_sample))

        # download gds from serverfiles
        try:
            makedirs(serverfiles.localpath(DOMAIN))
        except OSError:
            if path.exists(serverfiles.localpath(DOMAIN)):
                pass
            else:
                # There was an error on creation, so make sure we know about it
                raise
        gds_download(self.test_sample)

        # file in cache
        self.assertIsNone(gds_ensure_downloaded(self.test_sample))
        self.assertTrue(gds_is_cached(self.test_sample))

        gds = GDS(self.test_sample)
        self.assertIsNotNone(gds.info)
        self.assertEqual(gds.info['gene_count'], 9561)
        self.assertEqual(len(gds.info['samples']), 4)
        self.assertEqual(len(gds.info['subsets']), 2)

        self.assertEqual(gds.info['taxid'], self.test_organism)

        self.assertIsInstance(gds.get_data(), Table)
        self.assertIsInstance(gds.get_data(transpose=True), Table)


if __name__ == '__main__':
    unittest.main()



""" Differential Gene Expression """
import sys
import numpy as np
import pyqtgraph as pg
import scipy.special
import scipy.stats
import Orange.data

from types import SimpleNamespace as namespace
from AnyQt.QtGui import (
    QStandardItemModel, QPen
)
from AnyQt.QtCore import (
    Qt, pyqtSignal as Signal, pyqtSlot as Slot, QSize, QRectF, QLineF
)

from Orange.canvas.report import report
from Orange.preprocess import transformation
from Orange.widgets import widget, gui, settings
from Orange.widgets.utils import concurrent
from Orange.widgets.utils.datacaching import data_hints

from orangecontrib.bioinformatics.widgets.utils.settings import SetContextHandler
from orangecontrib.bioinformatics.widgets.utils import gui as guiutils
from orangecontrib.bioinformatics.widgets.utils.data import GENE_AS_ATTRIBUTE_NAME
from orangecontrib.bioinformatics.utils.statistics import score_hypergeometric_test


def score_fold_change(a, b, **kwargs):
    """
    Calculate the fold change between `a` and `b` samples.

    Parameters
    ----------
    a, b : array
        Arrays containing the samples
    axis : int
        Axis over which to compute the FC

    Returns
    -------
    FC : array
        The FC scores
    """
    axis = kwargs.get('axis', 0)

    mean_a = np.nanmean(a, axis=axis)
    mean_b = np.nanmean(b, axis=axis)
    res = mean_a / mean_b
    warning = None
    if np.any(res < 0):
        res[res < 0] = float("nan")
        warning = "Negative fold change scores were ignored. You should use another scoring method."
    return res, warning


def score_log_fold_change(a, b, **kwargs):
    """
    Return the log2(FC).

    See Also
    --------
    score_fold_change

    """
    axis = kwargs.get('axis', 0)
    s, w = score_fold_change(a, b, axis=axis)
    return np.log2(s), w


def score_ttest(a, b, **kwargs):
    axis = kwargs.get('axis', 0)
    T, P = scipy.stats.ttest_ind(a, b, axis=axis)
    return T, P


def score_ttest_t(a, b, **kwargs):
    axis = kwargs.get('axis', 0)
    T, _ = score_ttest(a, b, axis=axis)
    return T


def score_ttest_p(a, b, **kwargs):
    axis = kwargs.get('axis', 0)
    _, P = score_ttest(a, b, axis=axis)
    return P


def score_anova(*arrays, **kwargs):
    axis = kwargs.get('axis', 0)
    F, P = f_oneway(*arrays, axis=axis)
    return F, P


def score_anova_(*arrays, **kwargs):
    axis = kwargs.get('axis', 0)
    arrays = [np.asarray(arr, dtype=float) for arr in arrays]

    if not len(arrays) > 1:
        raise TypeError("Need at least 2 positional arguments")

    if not 0 <= axis < 2:
        raise ValueError("0 <= axis < 2")

    if not all(arrays[i].ndim == arrays[i + 1].ndim
               for i in range(len(arrays) - 2)):
        raise ValueError("All arrays must have the same number of dimensions")

    if axis >= arrays[0].ndim:
        raise ValueError()

    if axis == 0:
        arrays = [arr.T for arr in arrays]

    scores = [scipy.stats.f_oneway(*ars) for ars in zip(*arrays)]
    F, P = zip(*scores)
    return np.array(F, dtype=float), np.array(P, dtype=float)


def f_oneway(*arrays, **kwargs):
    """
    Perform a 1-way ANOVA

    Like `scipy.stats.f_oneway` but accept 2D arrays, with `axis`
    specifying over which axis to operate (in which axis the samples
    are stored).

    Parameters
    ----------
    A1, A2, ... : array_like
        The samples for each group.
    axis : int
        The axis which contain the samples.

    Returns
    -------
    F : array
        F scores
    P : array
        P values

    See also
    --------
    scipy.stats.f_oneway
    """
    axis = kwargs.get('axis', 0)
    arrays = [np.asarray(a, dtype=float) for a in arrays]
    alldata = np.concatenate(arrays, axis)
    bign = alldata.shape[axis]
    sstot = np.sum(alldata ** 2, axis) - (np.sum(alldata, axis) ** 2) / bign

    ssarrays = [(np.sum(a, axis, keepdims=True) ** 2) / a.shape[axis]
                for a in arrays]
    ssbn = np.sum(np.concatenate(ssarrays, axis), axis)
    ssbn -= (np.sum(alldata, axis) ** 2) / bign
    assert sstot.shape == ssbn.shape

    sswn = sstot - ssbn
    dfbn = len(arrays) - 1
    dfwn = bign - len(arrays)
    msb = ssbn / dfbn
    msw = sswn / dfwn
    f = msb / msw
    prob = scipy.special.fdtrc(dfbn, dfwn, f)
    return f, prob


def score_anova_f(*arrays, **kwargs):
    axis = kwargs.get('axis', 0)
    F, _ = score_anova(*arrays, axis=axis)
    return F


def score_anova_p(*arrays, **kwargs):
    axis = kwargs.get('axis', 0)
    _, P = score_anova(*arrays, axis=axis)
    return P


def score_signal_to_noise(a, b, **kwargs):
    axis = kwargs.get('axis', 0)
    mean_a = np.nanmean(a, axis=axis)
    mean_b = np.nanmean(b, axis=axis)

    std_a = np.nanstd(a, axis=axis, ddof=1)
    std_b = np.nanstd(b, axis=axis, ddof=1)

    return (mean_a - mean_b) / (std_a + std_b)


def score_mann_whitney(a, b, **kwargs):
    axis = kwargs.get('axis', 0)
    a, b = np.asarray(a, dtype=float), np.asarray(b, dtype=float)

    if not 0 <= axis < 2:
        raise ValueError("Axis")

    if a.ndim != b.ndim:
        raise ValueError

    if axis >= a.ndim:
        raise ValueError

    if axis == 0:
        a, b = a.T, b.T

    res = [scipy.stats.mannwhitneyu(a_, b_) for a_, b_ in zip(a, b)]
    U, P = zip(*res)
    return np.array(U), np.array(P)


def score_mann_whitney_u(a, b, **kwargs):
    axis = kwargs.get('axis', 0)
    U, _ = score_mann_whitney(a, b, axis=axis)
    return U


class InfiniteLine(pg.InfiniteLine):
    def paint(self, painter, option, widget=None):
        brect = self.boundingRect()
        c = brect.center()
        line = QLineF(brect.left(), c.y(), brect.right(), c.y())
        t = painter.transform()
        line = t.map(line)
        painter.save()
        painter.resetTransform()
        painter.setPen(self.currentPen)
        painter.drawLine(line)
        painter.restore()


def hypergeometric_test_score(*args, **kwargs):

    expression_treshold = kwargs.get('treshold', None)
    scores, _ = score_hypergeometric_test(*args, expression_treshold)
    return scores


class Histogram(pg.PlotWidget):
    """
    A histogram plot with interactive 'tail' selection
    """
    #: Emitted when the selection boundary has changed
    selectionChanged = Signal()
    #: Emitted when the selection boundary has been edited by the user
    #: (by dragging the boundary lines)
    selectionEdited = Signal()

    #: Selection mode
    NoSelection, Low, High, TwoSided, Middle = 0, 1, 2, 3, 4

    def __init__(self, parent=None, **kwargs):
        pg.PlotWidget.__init__(self, parent, **kwargs)

        self.getAxis("bottom").setLabel("Score")
        self.getAxis("left").setLabel("Counts")

        self.__data = None
        self.__histcurve = None

        self.__mode = Histogram.NoSelection
        self.__min = 0
        self.__max = 0

        def makeline(pos):
            pen = QPen(Qt.darkGray, 1)
            pen.setCosmetic(True)
            line = InfiniteLine(angle=90, pos=pos, pen=pen, movable=True)
            line.setCursor(Qt.SizeHorCursor)
            return line

        self.__cuthigh = makeline(self.__max)
        self.__cuthigh.sigPositionChanged.connect(self.__on_cuthigh_changed)
        self.__cuthigh.sigPositionChangeFinished.connect(self.selectionEdited)
        self.__cutlow = makeline(self.__min)
        self.__cutlow.sigPositionChanged.connect(self.__on_cutlow_changed)
        self.__cutlow.sigPositionChangeFinished.connect(self.selectionEdited)

        brush = pg.mkBrush((200, 200, 200, 180))
        self.__taillow = pg.PlotCurveItem(
            fillLevel=0, brush=brush, pen=QPen(Qt.NoPen))
        self.__taillow.setVisible(False)

        self.__tailhigh = pg.PlotCurveItem(
            fillLevel=0, brush=brush, pen=QPen(Qt.NoPen))
        self.__tailhigh.setVisible(False)

    def setData(self, hist, bins=None):
        """
        Set the histogram data
        """
        if bins is None:
            bins = np.arange(len(hist))

        self.__data = (hist, bins)
        if self.__histcurve is None:
            self.__histcurve = pg.PlotCurveItem(
                x=bins, y=hist, stepMode=True
            )
        else:
            self.__histcurve.setData(x=bins, y=hist, stepMode=True)

        self.__update()

    def setHistogramCurve(self, curveitem):
        """
        Set the histogram plot curve.
        """
        if self.__histcurve is curveitem:
            return

        if self.__histcurve is not None:
            self.removeItem(self.__histcurve)
            self.__histcurve = None
            self.__data = None

        if curveitem is not None:
            if not curveitem.opts["stepMode"]:
                raise ValueError("The curve must have `stepMode == True`")
            self.addItem(curveitem)
            self.__histcurve = curveitem
            self.__data = (curveitem.yData, curveitem.xData)

        self.__update()

    def histogramCurve(self):
        """
        Return the histogram plot curve.
        """
        return self.__histcurve

    def setSelectionMode(self, mode):
        """
        Set selection mode
        """
        if self.__mode != mode:
            self.__mode = mode
            self.__update_cutlines()
            self.__update_tails()

    def setLower(self, value):
        """
        Set the lower boundary value.
        """
        if self.__min != value:
            self.__min = value
            self.__update_cutlines()
            self.__update_tails()
            self.selectionChanged.emit()

    def setUpper(self, value):
        """
        Set the upper boundary value.
        """
        if self.__max != value:
            self.__max = value
            self.__update_cutlines()
            self.__update_tails()
            self.selectionChanged.emit()

    def setBoundary(self, lower, upper):
        """
        Set lower and upper boundary value.
        """
        changed = False
        if self.__min != lower:
            self.__min = lower
            changed = True

        if self.__max != upper:
            self.__max = upper
            changed = True

        if changed:
            self.__update_cutlines()
            self.__update_tails()
            self.selectionChanged.emit()

    def boundary(self):
        """
        Return the lower and upper boundary values.
        """
        return (self.__min, self.__max)

    def clear(self):
        """
        Clear the plot.
        """
        self.__data = None
        self.__histcurve = None
        super().clear()

    def __update(self):
        def additem(item):
            if item.scene() is not self.scene():
                self.addItem(item)

        def removeitem(item):
            if item.scene() is self.scene():
                self.removeItem(item)

        if self.__data is not None:
            additem(self.__cuthigh)
            additem(self.__cutlow)
            additem(self.__tailhigh)
            additem(self.__taillow)

            _, edges = self.__data
            # Update the allowable cutoff line bounds
            minx, maxx = np.min(edges), np.max(edges)
            span = maxx - minx
            bounds = minx - span * 0.005, maxx + span * 0.005

            self.__cuthigh.setBounds(bounds)
            self.__cutlow.setBounds(bounds)

            self.__update_cutlines()
            self.__update_tails()
        else:
            removeitem(self.__cuthigh)
            removeitem(self.__cutlow)
            removeitem(self.__tailhigh)
            removeitem(self.__taillow)

    def __update_cutlines(self):
        self.__cuthigh.setVisible(self.__mode & Histogram.High)
        self.__cuthigh.setValue(self.__max)
        self.__cutlow.setVisible(self.__mode & Histogram.Low)
        self.__cutlow.setValue(self.__min)

    def __update_tails(self):
        if self.__mode == Histogram.NoSelection:
            return
        if self.__data is None:
            return

        hist, edges = self.__data

        self.__taillow.setVisible(self.__mode & Histogram.Low)
        if self.__min > edges[0]:
            datalow = histogram_cut(hist, edges, edges[0], self.__min)
            self.__taillow.setData(*datalow, fillLevel=0, stepMode=True)
        else:
            self.__taillow.clear()

        self.__tailhigh.setVisible(self.__mode & Histogram.High)
        if self.__max < edges[-1]:
            datahigh = histogram_cut(hist, edges, self.__max, edges[-1])
            self.__tailhigh.setData(*datahigh, fillLevel=0, stepMode=True)
        else:
            self.__tailhigh.clear()

    def __on_cuthigh_changed(self):
        self.setUpper(self.__cuthigh.value())

    def __on_cutlow_changed(self):
        self.setLower(self.__cutlow.value())


def histogram_cut(hist, bins, low, high):
    """
    Return a subset of a histogram between low and high values.

    Parameters
    ----------
    hist : (N, ) array
        The histogram values/counts for each bin.
    bins : (N + 1) array
        The histogram bin edges.
    low, high : float
        The lower and upper edge where to cut the histogram

    Returns
    -------
    histsubset : (M, ) array
        The histogram subset
    binssubset : (M + 1) array
        New histogram bins. The first and the last value are equal
        to `low` and `high` respectively.

    Note that in general the first and the final bin widths are
    different then the widths in the input bins

    """
    if len(bins) < 2:
        raise ValueError()

    if low >= high:
        raise ValueError()

    low = max(bins[0], low)
    high = min(bins[-1], high)

    if low <= bins[0]:
        lowidx = 0
    else:
        lowidx = np.searchsorted(bins, low, side="left")

    if high >= bins[-1]:
        highidx = len(bins)
    else:
        highidx = np.searchsorted(bins, high, side="right")

    cbins = bins[lowidx: highidx]
    chist = hist[lowidx: highidx - 1]

    if cbins[0] > low:
        cbins = np.r_[low, cbins]
        chist = np.r_[hist[lowidx - 1], chist]

    if cbins[-1] < high:
        cbins = np.r_[cbins, high]
        chist = np.r_[chist, hist[highidx - 1]]

    assert cbins.size == chist.size + 1
    return cbins, chist


def test_low(array, low, high):
    return array <= low


def test_high(array, low, high):
    return array >= high


def test_two_tail(array, low, high):
    return (array >= high) | (array <= low)


def test_middle(array, low, high):
    return (array <= high) | (array >= low)


class OWDifferentialExpression(widget.OWWidget):
    name = "Differential Expression"
    description = "Gene selection by differential expression analysis."
    icon = "../widgets/icons/OWDifferentialExpression.svg"
    priority = 6

    inputs = [("Data", Orange.data.Table, "set_data")]
    outputs = [("Data subset", Orange.data.Table, widget.Default),
               ("Remaining data subset", Orange.data.Table),
               ("Selected genes", Orange.data.Table)]

    #: Selection types
    LowTail, HighTail, TwoTail = 1, 2, 3
    #: Test type - i.e a two sample (t-test, ...) or multi-sample (ANOVA) test
    TwoSampleTest, VarSampleTest = 1, 2
    #: Available scoring methods

    Scores = [
        ("Fold Change", TwoTail, TwoSampleTest, score_fold_change),
        ("log2(Fold Change)", TwoTail, TwoSampleTest, score_log_fold_change),
        ("T-test", TwoTail, TwoSampleTest, score_ttest_t),
        ("T-test P-value", LowTail, TwoSampleTest, score_ttest_p),
        ("ANOVA", HighTail, VarSampleTest, score_anova_f),
        ("ANOVA P-value", LowTail, VarSampleTest, score_anova_p),
        ("Signal to Noise Ratio", TwoTail, TwoSampleTest, score_signal_to_noise),
        ("Mann-Whitney", LowTail, TwoSampleTest, score_mann_whitney_u),
        ('Hypergeometric Test', TwoTail, TwoSampleTest, hypergeometric_test_score)
    ]

    settingsHandler = SetContextHandler()

    #: Selected score index.
    score_index = settings.Setting(0)
    #: Compute the null score distribution (label permutations).
    compute_null = settings.Setting(False)
    #: Number of permutations to for null score distribution.
    permutations_count = settings.Setting(20)
    #: Alpha value (significance) for the selection on background
    #: null score distribution.
    alpha_value = settings.Setting(0.01)
    #: N best for the fixed best N scores selection.
    n_best = settings.Setting(20)

    #: Stored thresholds for scores.
    thresholds = settings.Setting({
        "Fold Change": (0.5, 2.),
        "log2(Fold Change)": (-1, 1),
        "T-test": (-2, 2),
        "T-test P-value": (0.01, 0.01),
        "ANOVA": (0, 3),
        "ANOVA P-value": (0, 0.01),
    })

    auto_commit = settings.Setting(False)

    #: Current target group index
    current_group_index = settings.ContextSetting(-1)
    #: Stored (persistent) values selection for all target split groups.
    stored_selections = settings.ContextSetting([])

    graph_name = 'histogram.plotItem'

    def __init__(self, parent=None):
        widget.OWWidget.__init__(self, parent)

        self.min_value, self.max_value = \
            self.thresholds.get(self.Scores[self.score_index][0], (1, 0))

        # threshold defining expressed genes (for Hypergeometric Test)
        self.expression_threshold_value = 1.0

        #: Input data set
        self.data = None
        #: Current target group selection
        self.targets = []
        #: The computed scores
        self.scores = None
        #: The computed scores from label permutations
        self.nulldist = None

        self.__scores_future = self.__scores_state = None

        self.__in_progress = False

        self.test_f = {
            OWDifferentialExpression.LowTail: test_low,
            OWDifferentialExpression.HighTail: test_high,
            OWDifferentialExpression.TwoTail: test_two_tail,
        }

        self.histogram = Histogram(
            enableMouse=False, enableMenu=False, background="w"
        )
        self.histogram.enableAutoRange(enable=False)
        self.histogram.getPlotItem().hideButtons()  # hide auto range button
        self.histogram.getViewBox().setMouseEnabled(False, False)
        self.histogram.selectionChanged.connect(
            self.__on_histogram_plot_selection_changed
        )
        self.histogram.selectionEdited.connect(
            self._invalidate_selection
        )

        self.mainArea.layout().addWidget(self.histogram)

        box = gui.widgetBox(self.controlArea, "Info")

        self.dataInfoLabel = gui.widgetLabel(box, "No data on input.\n")
        self.dataInfoLabel.setWordWrap(True)
        self.selectedInfoLabel = gui.widgetLabel(box, "\n")

        box1 = gui.widgetBox(self.controlArea, "Scoring Method")
        gui.comboBox(box1, self, "score_index",
                     items=[sm[0] for sm in self.Scores],
                     callback=[self.on_scoring_method_changed, self.update_scores])

        self.expression_threshold_box = gui.widgetBox(self.controlArea, 'Expression threshold')
        self.expression_threshold = gui.doubleSpin(
            self.expression_threshold_box,
            self, "expression_threshold_value",
            minv=0, maxv=1e2, step=1e-2,
            callback=self.update_scores,
            callbackOnReturn=True)

        box = gui.widgetBox(self.controlArea, "Target Labels")
        self.label_selection_widget = guiutils.LabelSelectionWidget(self)
        self.label_selection_widget.setMaximumHeight(150)
        box.layout().addWidget(self.label_selection_widget)

        self.label_selection_widget.groupChanged.connect(
            self.on_label_activated)

        self.label_selection_widget.groupSelectionChanged.connect(
            self.on_target_changed)

        box = gui.widgetBox(self.controlArea, "Selection")
        box.layout().setSpacing(0)

        self.max_value_spin = gui.doubleSpin(
            box, self, "max_value", minv=-1e6, maxv=1e6, step=1e-6,
            label="Upper threshold:", callback=self.update_boundary,
            callbackOnReturn=True)

        self.low_value_spin = gui.doubleSpin(
            box, self, "min_value", minv=-1e6, maxv=1e6, step=1e-6,
            label="Lower threshold:", callback=self.update_boundary,
            callbackOnReturn=True)

        check = gui.checkBox(
            box, self, "compute_null", "Compute null distribution",
            callback=self.update_scores)

        perm_spin = gui.spin(
            box, self, "permutations_count", minv=1, maxv=50,
            label="Permutations:", callback=self.update_scores,
            callbackOnReturn=True)

        check.disables.append(perm_spin)

        box1 = gui.widgetBox(box, orientation='horizontal')

        pval_spin = gui.doubleSpin(
            box1, self, "alpha_value", minv=2e-7, maxv=1.0, step=1e-7,
            label="α-value:")
        pval_select = gui.button(
            box1, self, "Select", callback=self.select_p_best,
            autoDefault=False
        )
        check.disables.append(pval_spin)
        check.disables.append(pval_select)

        check.makeConsistent()

        box1 = gui.widgetBox(box, orientation='horizontal')
        gui.spin(box1, self, "n_best", 0, 10000, step=1,
                 label="Best Ranked:")
        gui.button(box1, self, "Select", callback=self.select_n_best,
                   autoDefault=False)

        box = gui.widgetBox(self.controlArea, "Output")

        acbox = gui.auto_commit(
            box, self, "auto_commit", "Commit", box=None)
        acbox.button.setDefault(True)

        gui.rubber(self.controlArea)

        self.on_scoring_method_changed()
        self._executor = concurrent.ThreadExecutor()

    def sizeHint(self):
        return QSize(800, 600)

    def clear(self):
        """Clear the widget state.
        """
        self.data = None
        self.targets = []
        self.stored_selections = []
        self.nulldist = None
        self.scores = None
        self.label_selection_widget.clear()
        self.clear_plot()
        self.dataInfoLabel.setText("No data on input.\n")
        self.selectedInfoLabel.setText("\n")
        self.__cancel_pending()

    def clear_plot(self):
        """Clear the histogram plot.
        """
        self.histogram.clear()

    def initialize(self, data):
        """Initialize widget state from the data."""

        col_targets, row_targets = guiutils.group_candidates(data)
        modelitems = [guiutils.standarditem_from(obj)
                      for obj in col_targets + row_targets]

        model = QStandardItemModel()
        for item in modelitems:
            model.appendRow(item)

        self.label_selection_widget.setModel(model)

        self.targets = col_targets + row_targets
        # Default selections for all group keys
        # (the first value is selected)
        self.stored_selections = [[0] for _ in self.targets]

    def selected_split(self):
        index = self.label_selection_widget.currentGroupIndex()
        if not (0 <= index < len(self.targets)):
            return None, ()

        grp = self.targets[index]
        selection = self.label_selection_widget.currentGroupSelection()
        selected_indices = [ind.row() for ind in selection.indexes()]
        return grp, selected_indices

    def set_data(self, data):
        self.closeContext()

        self.clear()
        self.error([0, 1])
        self.data = data

        if self.data is not None:
            self.initialize(data)

        if self.data is not None and not self.targets:
            # If both attr. labels and classes are missing, show an error
            self.error(
                1, "Cannot compute gene scores! Differential expression "
                   "widget requires a data-set with a discrete class "
                   "variable(s) or column labels!"
            )
            self.data = None

        if self.data is not None:
            # Initialize the selected groups/labels.
            # Default selected group key
            index = 0
            rowshint = data_hints.get_hint(data, GENE_AS_ATTRIBUTE_NAME, False)

            if not rowshint:
                # Select the first row group split candidate (if available)
                indices = [i for i, grp in enumerate(self.targets)
                           if isinstance(grp, guiutils.RowGroup)]
                if indices:
                    index = indices[0]

            self.current_group_index = index

            # Restore target label selection from context settings
            items = {(grp.name, val)
                     for grp in self.targets for val in grp.values}
            self.openContext(items)

            # Restore current group / selection
            model = self.label_selection_widget.model()
            selection = [model.index(i, 0, model.index(keyind, 0))
                         for keyind, selection in enumerate(self.stored_selections)
                         for i in selection]
            selection = guiutils.itemselection(selection)

            self.label_selection_widget.setSelection(selection)
            self.label_selection_widget.setCurrentGroupIndex(
                self.current_group_index)

        self.commit()

    def update_scores(self):
        """Compute the scores and update the histogram.
        """
        self.__cancel_pending()
        self.clear_plot()
        self.scores = None
        self.nulldist = None
        self.error(0)

        grp, split_selection = self.selected_split()

        if not self.data or grp is None:
            return

        _, side, test_type, score_func = self.Scores[self.score_index]

        def compute_scores(X, group_indices, warn=False):
            arrays = [X[ind] for ind in group_indices]
            ss = score_func(*arrays,
                            axis=0,
                            treshold=self.expression_threshold_value)

            return ss[0] if isinstance(ss, tuple) and not warn else ss

        def permute_indices(group_indices, random_state=None):
            assert all(ind.dtype.kind == "i" for ind in group_indices)
            assert all(ind.ndim == 1 for ind in group_indices)
            if random_state is None:
                random_state = np.random
            joined = np.hstack(group_indices)
            random_state.shuffle(joined)
            split_ind = np.cumsum([len(ind) for ind in group_indices])
            return np.split(joined, split_ind[:-1])

        if isinstance(grp, guiutils.RowGroup):
            axis = 0
        else:
            axis = 1

        if test_type == OWDifferentialExpression.TwoSampleTest:
            G1 = guiutils.group_selection_mask(
                self.data, grp, split_selection)
            G2 = ~G1
            indices = [np.flatnonzero(G1), np.flatnonzero(G2)]

        elif test_type == OWDifferentialExpression.VarSampleTest:
            indices = [guiutils.group_selection_mask(self.data, grp, [i])
                       for i in range(len(grp.values))]
            indices = [np.flatnonzero(ind) for ind in indices]
        else:
            assert False

        if not all(ind.size > 0 for ind in indices):
            self.error(0, "Target labels most exclude/include at least one "
                          "value.")
            self.scores = None
            self.nulldist = None
            self.update_data_info_label()
            return

        X = self.data.X
        if axis == 1:
            X = X.T

        # TODO: Check that each label has more than one measurement,
        # raise warning otherwise.

        def compute_scores_with_perm(X, indices, nperm=0, rstate=None,
                                     progress_advance=None):
            warning = None
            scores = compute_scores(X, indices, warn=True)
            if isinstance(scores, tuple):
                scores, warning = scores

            if progress_advance is not None:
                progress_advance()
            null_scores = []
            if nperm > 0:
                if rstate is None:
                    rstate = np.random.RandomState(0)

                for i in range(nperm):
                    p_indices = permute_indices(indices, rstate)
                    assert all(pind.shape == ind.shape
                               for pind, ind in zip(indices, p_indices))
                    pscore = compute_scores(X, p_indices)
                    assert pscore.shape == scores.shape
                    null_scores.append(pscore)
                    if progress_advance is not None:
                        progress_advance()

            return scores, null_scores, warning

        p_advance = concurrent.methodinvoke(
            self, "progressBarAdvance", (float,))
        state = namespace(cancelled=False, advance=p_advance)

        def progress():
            if state.cancelled:
                raise concurrent.CancelledError
            else:
                state.advance(100 / (nperm + 1))

        self.progressBarInit()
        set_scores = concurrent.methodinvoke(
            self, "__set_score_results", (concurrent.Future,))

        nperm = self.permutations_count if self.compute_null else 0
        self.__scores_state = state
        self.__scores_future = self._executor.submit(
                compute_scores_with_perm, X, indices, nperm,
                progress_advance=progress)
        self.__scores_future.add_done_callback(set_scores)

    @Slot(float)
    def __pb_advance(self, value):
        self.progressBarAdvance(value, )

    @Slot(float)
    def progressBarAdvance(self, value):
        if not self.__in_progress:
            self.__in_progress = True
            try:
                super().progressBarAdvance(value)
            finally:
                self.__in_progress = False

    @Slot(concurrent.Future)
    def __set_score_results(self, scores):
        # set score results from a Future
        self.error(1)
        self.warning(10)
        if scores is self.__scores_future:
            self.histogram.setUpdatesEnabled(True)
            self.progressBarFinished()

            if not self.__scores_state.cancelled:
                try:
                    results = scores.result()
                except Exception as ex:
                    self.error(1, "Error: {!s}".format(ex))
                else:
                    self.set_scores(*results)

        elif self.__scores_future is None:
            self.histogram.setUpdatesEnabled(True)
            self.progressBarFinished()

    def __cancel_pending(self):
        if self.__scores_future is not None:
            self.__scores_future.cancel()
            self.__scores_state.cancelled = True
            self.__scores_state = self.__scores_future = None

    def set_scores(self, scores, null_scores=None, warning=None):
        self.scores = scores
        self.nulldist = null_scores

        if null_scores:
            nulldist = np.array(null_scores, dtype=float)
        else:
            nulldist = None

        self.warning(10, warning)

        self.setup_plot(self.score_index, scores, nulldist)
        self.update_data_info_label()
        self.update_selected_info_label()
        self.commit()

    def setup_plot(self, scoreindex, scores, nulldist=None):
        """
        Setup the score histogram plot

        Parameters
        ----------
        scoreindex : int
            Score index (into OWFeatureSelection.Scores)
        scores : (N, ) array
            The scores obtained
        nulldist (P, N) array optional
            The scores obtained under P permutations of labels.
        """
        score_name, side, test_type, _ = self.Scores[scoreindex]
        low, high = self.thresholds.get(score_name, (-np.inf, np.inf))

        validmask = np.isfinite(scores)
        validscores = scores[validmask]

        nbins = int(max(np.ceil(np.sqrt(len(validscores))), 20))
        freq, edges = np.histogram(validscores, bins=nbins)
        self.histogram.setHistogramCurve(
            pg.PlotCurveItem(x=edges, y=freq, stepMode=True,
                             pen=pg.mkPen("b", width=2))
        )

        if nulldist is not None:
            nulldist = nulldist.ravel()
            validmask = np.isfinite(nulldist)
            validnulldist = nulldist[validmask]
            nullbins = edges  # XXX: extend to the full range of nulldist
            nullfreq, _ = np.histogram(validnulldist, bins=nullbins)
            nullfreq = nullfreq * (freq.sum() / nullfreq.sum())
            nullitem = pg.PlotCurveItem(
                x=nullbins, y=nullfreq, stepMode=True,
                pen=pg.mkPen((50, 50, 50, 100))
            )
            # Ensure it stacks behind the main curve
            nullitem.setZValue(nullitem.zValue() - 10)
            self.histogram.addItem(nullitem)

        # Restore saved thresholds
        eps = np.finfo(float).eps
        minx, maxx = edges[0] - eps, edges[-1] + eps
        low, high = max(low, minx), min(high, maxx)

        if side == OWDifferentialExpression.LowTail:
            mode = Histogram.Low
        elif side == OWDifferentialExpression.HighTail:
            mode = Histogram.High
        elif side == OWDifferentialExpression.TwoTail:
            mode = Histogram.TwoSided
        else:
            assert False
        self.histogram.setSelectionMode(mode)
        self.histogram.setBoundary(low, high)

        # If this is a two sample test add markers to the left and right
        # plot indicating which group is over-expressed in that part
        if test_type == OWDifferentialExpression.TwoSampleTest and \
                side == OWDifferentialExpression.TwoTail:
            maxy = np.max(freq)
            # XXX: Change use of integer constant
            if scoreindex == 0:  # fold change is centered on 1.0
                x1, y1 = (minx + 1) / 2, maxy
                x2, y2 = (maxx + 1) / 2, maxy
            else:
                x1, y1 = minx / 2, maxy
                x2, y2 = maxx / 2, maxy

            grp, selected_indices = self.selected_split()

            values = grp.values
            selected_values = [values[i] for i in selected_indices]

            left = ", ".join(v for v in values if v not in selected_values)
            right = ", ".join(v for v in selected_values)

            labelitem = pg.TextItem(left, color=(40, 40, 40))
            labelitem.setPos(x1, y1)
            self.histogram.addItem(labelitem)

            labelitem = pg.TextItem(right, color=(40, 40, 40))
            labelitem.setPos(x2, y2)
            self.histogram.addItem(labelitem)

        miny, maxy = 0.0, np.max(freq)
        if nulldist is not None:
            maxy = max(np.max(nullfreq), maxy)

        if not np.any(np.isnan([maxx, minx, maxy])):
            self.histogram.setRange(
                QRectF(minx, miny, maxx - minx, maxy - miny),
                padding=0.05
            )

    def update_data_info_label(self):
        if self.data is not None:
            samples, genes = len(self.data), len(self.data.domain.attributes)
            grp, indices = self.selected_split()

            if isinstance(grp, guiutils.ColumnGroup):
                samples, genes = genes, samples

            target_labels = [grp.values[i] for i in indices]
            text = "%i samples, %i genes\n" % (samples, genes)
            text += "Sample target: '%s'" % (",".join(target_labels))
        else:
            text = "No data on input.\n"

        self.dataInfoLabel.setText(text)

    def update_selected_info_label(self):
        pl = lambda c: "" if c == 1 else "s"
        if self.data is not None and self.scores is not None:
            scores = self.scores
            low, high = self.min_value, self.max_value
            _, side, _, _ = self.Scores[self.score_index]
            test = self.test_f[side]
            count_undef = np.count_nonzero(np.isnan(scores))
            count_scores = len(scores)
            scores = scores[np.isfinite(scores)]

            nselected = np.count_nonzero(test(scores, low, high))
            defined_txt = ("{} of {} score{} undefined."
                           .format(count_undef, count_scores, pl(count_scores)))

        elif self.data is not None:
            nselected = 0
            defined_txt = "No defined scores"
        else:
            nselected = 0
            defined_txt = ""

        self.selectedInfoLabel.setText(
            defined_txt + "\n" +
            "{} selected gene{}".format(nselected, pl(nselected))
        )

    def __on_histogram_plot_selection_changed(self):
        low, high = self.histogram.boundary()
        scorename, side, _, _ = self.Scores[self.score_index]
        self.thresholds[scorename] = (low, high)
        self.min_value = low
        self.max_value = high
        self.update_selected_info_label()

    def update_boundary(self):
        # The cutoff boundary value has been changed by the user
        # (in the controlArea widgets). Update the histogram plot
        # accordingly.
        if self.data is None:
            return

        _, side, _, _ = self.Scores[self.score_index]
        if side == OWDifferentialExpression.LowTail:
            self.histogram.setLower(self.min_value)
        elif side == OWDifferentialExpression.HighTail:
            self.histogram.setUpper(self.max_value)
        elif side == OWDifferentialExpression.TwoTail:
            self.histogram.setBoundary(self.min_value, self.max_value)

        self._invalidate_selection()

    def select_n_best(self):
        """
        Select the `self.n_best` scored genes.
        """
        if self.scores is None:
            return

        score_name, side, _, _ = self.Scores[self.score_index]
        scores = self.scores
        scores = np.sort(scores[np.isfinite(scores)])

        if side == OWDifferentialExpression.HighTail:
            cut = scores[-np.clip(self.n_best, 1, len(scores))]
            self.histogram.setUpper(cut)
        elif side == OWDifferentialExpression.LowTail:
            cut = scores[np.clip(self.n_best, 0, len(scores) - 1)]
            self.histogram.setLower(cut)
        elif side == OWDifferentialExpression.TwoTail:
            n = min(self.n_best, len(scores))
            scoresabs = np.abs(scores)
            if score_name == "Fold Change":
                # comparing fold change on a logarithmic scale
                scores = np.log2(scoresabs)
                scores = scores[np.isfinite(scoresabs)]
            scoresabs = np.sort(np.abs(scores))
            limit = (scoresabs[-n] + scoresabs[-min(n+1, len(scores))]) / 2
            cuthigh, cutlow = limit, -limit
            if score_name == "Fold Change":
                cuthigh, cutlow = 2**cuthigh, 2**cutlow
            self.histogram.setBoundary(cutlow, cuthigh)
        self._invalidate_selection()

    def select_p_best(self):
        if not self.nulldist:
            return

        _, side, _, _ = self.Scores[self.score_index]
        nulldist = np.asarray(self.nulldist).ravel()
        nulldist = nulldist[np.isfinite(nulldist)]
        nulldist = np.sort(nulldist)

        assert 0 <= self.alpha_value <= 1
        p = self.alpha_value
        if side == OWDifferentialExpression.HighTail:
            cut = np.percentile(nulldist, 100 * (1 - p))
            self.max_value = cut
            self.histogram.setUpper(cut)
        elif side == OWDifferentialExpression.LowTail:
            cut = np.percentile(nulldist, 100 * p)
            self.min_value = cut
            self.histogram.setLower(cut)
        elif side == OWDifferentialExpression.TwoTail:
            p1, p2 = np.percentile(nulldist, [100 * p / 2, 100 * (1 - p / 2)])
            self.histogram.setBoundary(p1, p2)
        self._invalidate_selection()

    def _invalidate_selection(self):
        self.commit()

    def on_target_changed(self):
        grp, indices = self.selected_split()
        if grp is None:
            return
        # Store target group label selection.
        self.stored_selections[self.targets.index(grp)] = indices
        self.update_scores()

    def on_label_activated(self, index):
        self.current_group_index = index
        self.update_scores()

    def on_scoring_method_changed(self):
        _, _, test_type, score_method = self.Scores[self.score_index]
        self.label_selection_widget.values_view.setEnabled(
            test_type == OWDifferentialExpression.TwoSampleTest
        )

        # TODO: refactor this, not pretty :)
        self.expression_threshold_box.setHidden(score_method.__name__ != hypergeometric_test_score.__name__)

        self.__update_threshold_spinbox()

    def __update_threshold_spinbox(self):
        _, side, _, _ = self.Scores[self.score_index]
        self.low_value_spin.setVisible(side & OWDifferentialExpression.LowTail)
        self.max_value_spin.setVisible(side & OWDifferentialExpression.HighTail)

    def commit(self):
        """
        Commit (send) the outputs.
        """
        if self.data is None or self.scores is None:
            return

        grp, _ = self.selected_split()
        if isinstance(grp, guiutils.RowGroup):
            axis = 1
        else:
            axis = 0

        score_name, side, _, _ = self.Scores[self.score_index]
        low, high = self.histogram.boundary()

        scores = self.scores
        mask = np.isfinite(scores)
        test = self.test_f[side]
        selected_masked = test(scores[mask], low, high)

        selected = np.zeros_like(scores, dtype=bool)
        selected[mask] = selected_masked

        indices = np.flatnonzero(selected)
        remaining = np.flatnonzero(~selected)

        domain = self.data.domain

        if axis == 0:
            # Select rows
            score_var = Orange.data.ContinuousVariable(score_name)
            domain_selected_genes = Orange.data.Domain([], metas=domain.metas + (score_var,))

            domain = Orange.data.Domain(domain.attributes, domain.class_vars, domain.metas + (score_var,))

            data = self.data.from_table(domain, self.data)
            table_selected_genes = self.data.from_table(domain_selected_genes, self.data)

            data[:, score_var] = np.c_[scores]
            table_selected_genes[:, score_var] = np.c_[scores]

            subsetdata = data[indices]
            remainingdata = data[remaining]

            self.send("Selected genes", table_selected_genes[indices])
        else:
            domain_selected_genes = Orange.data.Domain([], metas=[Orange.data.StringVariable('genes'),
                                                                  Orange.data.ContinuousVariable(score_name)])
            data_selected_genes = []

            # select columns
            attrs = [copy_variable(var) for var in domain.attributes]
            for var, score in zip(attrs, scores):
                var.attributes[score_name] = str(score)
                data_selected_genes.append([var.name, score])

            table_selected_genes = Orange.data.Table(domain_selected_genes, data_selected_genes)

            selected_attrs = [attrs[i] for i in indices]
            remaining_attrs = [attrs[i] for i in remaining]

            domain = Orange.data.Domain(
                selected_attrs, domain.class_vars, domain.metas)
            subsetdata = self.data.from_table(domain, self.data)

            domain = Orange.data.Domain(
                remaining_attrs, domain.class_vars, domain.metas)
            remainingdata = self.data.from_table(domain, self.data)

            self.send("Selected genes", table_selected_genes[indices])

        self.send("Data subset", subsetdata)
        self.send("Remaining data subset", remainingdata)

    def send_report(self):
        self.report_plot()
        caption = report.render_items_vert((
            ("Scoring method", self.Scores[self.score_index][0]),
            ("Upper treshold", self.max_value),
            ("Lower threshold", self.min_value),
            ("Compute null distribution", self.compute_null),
            ("Permutations", self.permutations_count),
            ("α-value", self.alpha_value),
            ("Best Ranked", self.n_best)
        ))
        self.report_caption(caption)
        self.report_caption(self.selectedInfoLabel.text())

    def onDeleteWidget(self):
        super().onDeleteWidget()
        self.clear()
        self.__cancel_pending()
        self._executor.shutdown(wait=True)


def copy_variable(var):
    clone = var.copy(compute_value=transformation.Identity(var))
    clone.attributes = dict(var.attributes)
    return clone


if __name__ == "__main__":

    def main(argv=None):
        from AnyQt.QtWidgets import QApplication
        app = QApplication(list(argv) if argv else [])
        argv = app.arguments()
        if len(argv) > 1:
            filename = argv[1]
        else:
            filename = "geo-gds360"
        data = Orange.data.Table(filename)

        w = OWDifferentialExpression()
        w.show()
        w.raise_()
        w.set_data(data)
        rval = app.exec_()
        w.set_data(None)
        w.saveSettings()
        w.onDeleteWidget()
        return rval

    sys.exit(main(sys.argv))



""" GeneNameMatching """
import threading
import numpy as np
import re


from typing import Set, List

from AnyQt.QtWidgets import (
    QSplitter, QTableView,  QHeaderView, QAbstractItemView, QStyle
)
from AnyQt.QtCore import (
    Qt, QSize, QThreadPool, QAbstractTableModel, QVariant, QModelIndex
)
from AnyQt.QtGui import (
    QFont, QColor
)

from Orange.widgets.gui import (
    vBox, comboBox, ProgressBar, widgetBox, auto_commit, widgetLabel, checkBox,
    rubber, lineEdit, LinkRole, LinkStyledItemDelegate
)
from Orange.widgets.widget import OWWidget
from Orange.widgets.utils import itemmodels
from Orange.widgets.settings import Setting
from Orange.widgets.utils.signals import Output, Input
from Orange.data import StringVariable, DiscreteVariable, Domain, Table, filter as table_filter


from orangecontrib.bioinformatics.widgets.utils.data import (
    TAX_ID, GENE_AS_ATTRIBUTE_NAME, GENE_ID_COLUMN, GENE_ID_ATTRIBUTE
)
from orangecontrib.bioinformatics.widgets.utils.concurrent import Worker
from orangecontrib.bioinformatics.ncbi import taxonomy
from orangecontrib.bioinformatics.ncbi.gene import GeneMatcher, Gene, NCBI_ID, GENE_MATCHER_HEADER, NCBI_DETAIL_LINK

from functools import lru_cache


class GeneInfoModel(itemmodels.PyTableModel):
    def __init__(self,  *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.header_labels, self.gene_attributes = GENE_MATCHER_HEADER
        self.setHorizontalHeaderLabels(self.header_labels)

        try:
            # note: make sure ncbi_id is set in GENE_MATCHER_HEADER
            self.entrez_column_index = self.gene_attributes.index('ncbi_id')
        except ValueError as e:
            raise ValueError("Make sure 'ncbi_id' is set in gene.GENE_MATCHER_HEADER")

        self.genes = None
        self.data_table = None

        self.font = QFont()
        self.font.setUnderline(True)
        self.color = QColor(Qt.blue)

        @lru_cache(maxsize=10000)
        def _row_instance(row, column):
            return self[int(row)][int(column)]
        self._row_instance = _row_instance

    def initialize(self, list_of_genes):
        self.genes = list_of_genes
        self.__table_from_genes([gene for gene in list_of_genes if gene.ncbi_id])
        self.show_table()

    def columnCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else self._table.shape[1]

    def clear(self):
        self.beginResetModel()
        self._table = np.array([[]])
        self.resetSorting()
        self._roleData.clear()
        self.endResetModel()

    def data(self, index, role,
             _str=str,
             _Qt_DisplayRole=Qt.DisplayRole,
             _Qt_EditRole=Qt.EditRole,
             _Qt_FontRole=Qt.FontRole,
             _Qt_ForegroundRole=Qt.ForegroundRole,
             _LinkRolee=LinkRole,
             _recognizedRoles=frozenset([Qt.DisplayRole, Qt.EditRole, Qt.FontRole, Qt.ForegroundRole, LinkRole])):

        if role not in _recognizedRoles:
            return None

        row, col = index.row(), index.column()
        if not 0 <= row <= self.rowCount():
            return None
        row = self.mapToSourceRows(row)

        try:
            # value = self[row][col]
            value = self._row_instance(row, col)
        except IndexError:
            return

        if role == Qt.DisplayRole:
            return QVariant(str(value))
        elif role == Qt.ToolTipRole:
            return QVariant(str(value))

        if col == self.entrez_column_index:
            if role == _Qt_ForegroundRole:
                return self.color
            elif role == _Qt_FontRole:
                return self.font
            elif role == _LinkRolee:
                return NCBI_DETAIL_LINK.format(value)

    def __table_from_genes(self, list_of_genes):
        # type: (list) -> None
        self.data_table = np.asarray([gene.to_list() for gene in list_of_genes])

    def filter_table(self, filter_pattern):
        _, col_size = self.data_table.shape
        invalid_result = np.array([-1 for _ in range(col_size)])

        filtered_rows = []
        for row in self.data_table:
            match_result = np.core.defchararray.rfind(row.astype(str), filter_pattern)
            filtered_rows.append(not np.array_equal(match_result, invalid_result))
        return filtered_rows

    def show_table(self, filter_pattern=None):
        # Don't call filter if filter_pattern is empty
        if filter_pattern:
            # clear cache if model changes
            self._row_instance.cache_clear()
            self.wrap(self.data_table[self.filter_table(filter_pattern)])
        else:
            self.wrap(self.data_table)


class UnknownGeneInfoModel(itemmodels.PyListModel):
    def __init__(self,  *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.header_labels = ['IDs from the input data without corresponding Entrez ID']
        self.genes = []

    def initialize(self, list_of_genes):
        self.genes = list_of_genes
        self.wrap([', '.join([gene.input_name for gene in list_of_genes if not gene.ncbi_id])])

    def data(self, index, role=Qt.DisplayRole):
        row = index.row()
        if role in [self.list_item_role, Qt.EditRole] and self._is_index_valid(index):
            return self[row]
        elif role == Qt.TextAlignmentRole:
            return Qt.AlignLeft | Qt.AlignTop
        elif self._is_index_valid(row):
            return self._other_data[row].get(role, None)

    def headerData(self, section, orientation, role=Qt.DisplayRole):

        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self.header_labels[section]
        return QAbstractTableModel.headerData(self, section, orientation, role)


class OWGeneNameMatcher(OWWidget):
    name = "Gene Name Matcher"
    description = "Tool for working with genes"
    icon = "../widgets/icons/OWGeneInfo.svg"
    priority = 5
    want_main_area = True

    use_attr_names = Setting(True)
    selected_organism = Setting(11)

    search_pattern = Setting('')
    gene_as_attr_name = Setting(0)
    exclude_unmatched = Setting(True)
    include_entrez_id = Setting(True)
    replace_id_with_symbol = Setting(True)

    auto_commit = Setting(True)

    class Inputs:
        data_table = Input("Data", Table)

    class Outputs:
        custom_data_table = Output("Data", Table)
        gene_matcher_results = Output("Genes", Table)

    class Information(OWWidget.Information):
        pass

    def sizeHint(self):
        return QSize(1280, 960)

    def __init__(self):
        super().__init__()
        # ATTRIBUTES #

        # input data
        self.input_data = None
        self.input_genes = None
        self.tax_id = None
        self.column_candidates = []
        self.selected_gene_col = None

        # input options
        self.organisms = []

        # gene matcher
        self.gene_matcher = None

        # output data
        self.output_data_table = None
        self.gene_id_column = None
        self.gene_id_attribute = None

        # threads
        self.threadpool = QThreadPool(self)
        self.workers = None

        # progress bar
        self.progress_bar = None

        # GUI SECTION #

        # Control area
        self.info_box = widgetLabel(
            widgetBox(self.controlArea, "Info", addSpace=True), 'No data on input.\n'
        )

        organism_box = vBox(self.controlArea, 'Organism')
        self.organism_select_combobox = comboBox(organism_box, self,
                                                 'selected_organism',
                                                 callback=self.on_input_option_change)

        self.get_available_organisms()
        self.organism_select_combobox.setCurrentIndex(self.selected_organism)

        box = widgetBox(self.controlArea, 'Gene IDs in the input data')
        self.gene_columns_model = itemmodels.DomainModel(valid_types=(StringVariable, DiscreteVariable))
        self.gene_column_combobox = comboBox(box, self, 'selected_gene_col',
                                             label='Stored in data column',
                                             model=self.gene_columns_model,
                                             sendSelectedValue=True,
                                             callback=self.on_input_option_change)

        self.attr_names_checkbox = checkBox(box, self, 'use_attr_names', 'Stored as feature (column) names',
                                            disables=[(-1, self.gene_column_combobox)],
                                            callback=self.on_input_option_change)

        self.gene_column_combobox.setDisabled(bool(self.use_attr_names))

        output_box = vBox(self.controlArea, 'Output')

        # separator(output_box)
        # output_box.layout().addWidget(horizontal_line())
        # separator(output_box)
        self.exclude_radio = checkBox(output_box, self,
                                      'exclude_unmatched',
                                      'Exclude unmatched genes',
                                      callback=self.on_output_option_change)

        self.replace_radio = checkBox(output_box, self,
                                      'replace_id_with_symbol',
                                      'Replace feature IDs with gene names',
                                      callback=self.on_output_option_change)

        auto_commit(self.controlArea, self, "auto_commit", "&Commit", box=False)

        rubber(self.controlArea)

        # Main area
        self.filter = lineEdit(self.mainArea, self,
                               'search_pattern', 'Filter:',
                               callbackOnType=True, callback=self.apply_filter)
        # rubber(self.radio_group)
        self.mainArea.layout().addWidget(self.filter)

        # set splitter
        self.splitter = QSplitter()
        self.splitter.setOrientation(Qt.Vertical)

        self.table_model = GeneInfoModel()
        self.table_view = QTableView()
        self.table_view.setAlternatingRowColors(True)
        self.table_view.viewport().setMouseTracking(True)
        self.table_view.setSortingEnabled(True)
        self.table_view.setShowGrid(False)
        self.table_view.verticalHeader().hide()
        # self.table_view.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)

        self.unknown_model = UnknownGeneInfoModel()

        self.unknown_view = QTableView()
        self.unknown_view.setModel(self.unknown_model)
        self.unknown_view.verticalHeader().hide()
        self.unknown_view.setShowGrid(False)
        self.unknown_view.setSelectionMode(QAbstractItemView.NoSelection)
        self.unknown_view.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)

        self.splitter.addWidget(self.table_view)
        self.splitter.addWidget(self.unknown_view)

        self.splitter.setStretchFactor(0, 90)
        self.splitter.setStretchFactor(1, 10)

        self.mainArea.layout().addWidget(self.splitter)

    def apply_filter(self):
        # filter only if input data is present and model is populated
        if self.input_data is not None and self.table_model.data_table is not None:
            self.table_view.clearSpans()
            self.table_view.setModel(None)
            self.table_view.setSortingEnabled(False)
            self.table_model.show_table(str(self.search_pattern))
            self.table_view.setModel(self.table_model)
            self.table_view.selectionModel().selectionChanged.connect(self.invalidate)
            self.table_view.setSortingEnabled(True)

    def __reset_widget_state(self):
        self.Outputs.custom_data_table.send(None)
        self.Outputs.gene_matcher_results.send(None)
        self.table_view.clearSpans()
        self.table_view.setModel(None)
        self.table_model.clear()
        self.unknown_model.clear()
        self._update_info_box()

    def __selection_changed(self):
        genes = [model_index.data() for model_index in self.extended_view.get_selected_gens()]
        self.extended_view.set_info_model(genes)

    def _update_info_box(self):

        if self.input_genes and self.gene_matcher:
            num_genes = len(self.gene_matcher.genes)
            known_genes = len(self.gene_matcher.get_known_genes())

            info_text = '{} genes in input data\n' \
                        '{} genes match Entrez database\n' \
                        '{} genes with match conflicts\n'.format(num_genes, known_genes, num_genes - known_genes)

        else:
            info_text = 'No data on input.'

        self.info_box.setText(info_text)

    def _progress_advance(self):
        # GUI should be updated in main thread. That's why we are calling advance method here
        if self.progress_bar:
            self.progress_bar.advance()

    def _handle_matcher_results(self):
        assert threading.current_thread() == threading.main_thread()

        if self.progress_bar:
            self.progress_bar.finish()
            self.setStatusMessage('')

        # update info box
        self._update_info_box()

        # set output options
        self.toggle_radio_options()

        # if no known genes, clean up and return
        if not len(self.gene_matcher.get_known_genes()):
            self.table_model.wrap(np.array([[]]))
            self.__reset_widget_state()
            return

        # set known genes
        self.table_model.initialize(self.gene_matcher.genes)
        self.table_view.setModel(self.table_model)
        self.table_view.selectionModel().selectionChanged.connect(self.invalidate)
        self.table_view.setSelectionBehavior(QAbstractItemView.SelectRows)

        self.table_view.setItemDelegateForColumn(
            self.table_model.entrez_column_index, LinkStyledItemDelegate(self.table_view)
        )
        v_header = self.table_view.verticalHeader()
        option = self.table_view.viewOptions()
        size = self.table_view.style().sizeFromContents(
            QStyle.CT_ItemViewItem, option,
            QSize(20, 20), self.table_view)

        v_header.setDefaultSectionSize(size.height() + 2)
        v_header.setMinimumSectionSize(5)
        self.table_view.horizontalHeader().setStretchLastSection(True)

        # set unknown genes
        self.unknown_model.initialize(self.gene_matcher.genes)
        self.unknown_view.verticalHeader().setStretchLastSection(True)

        self.on_output_option_change()

    def get_available_organisms(self):
        available_organism = sorted([(tax_id, taxonomy.name(tax_id)) for tax_id in taxonomy.common_taxids()],
                                    key=lambda x: x[1])

        self.organisms = [tax_id[0] for tax_id in available_organism]
        self.organism_select_combobox.addItems([tax_id[1] for tax_id in available_organism])

    def gene_names_from_table(self):
        """ Extract and return gene names from `Orange.data.Table`.
        """
        self.input_genes = []
        if self.input_data:
            if self.use_attr_names:
                self.input_genes = [str(attr.name).strip() for attr in self.input_data.domain.attributes]
            elif self.selected_gene_col:
                if self.selected_gene_col in self.input_data.domain:
                    self.input_genes = [str(e[self.selected_gene_col]) for e in self.input_data
                                        if not np.isnan(e[self.selected_gene_col])]

    def _update_gene_matcher(self):
        self.gene_names_from_table()

        if not self.input_genes:
            self._update_info_box()

        if not self.gene_matcher:
            self.gene_matcher = GeneMatcher(self.get_selected_organism(), case_insensitive=True)

        self.gene_matcher.genes = self.input_genes
        self.gene_matcher.organism = self.get_selected_organism()

    def get_selected_organism(self):
        return self.organisms[self.selected_organism]

    def match_genes(self):
        if self.gene_matcher:
            # init progress bar
            self.progress_bar = ProgressBar(self, iterations=len(self.gene_matcher.genes))
            # status message
            self.setStatusMessage('Gene matcher running')

            worker = Worker(self.gene_matcher.run_matcher, progress_callback=True)
            worker.signals.progress.connect(self._progress_advance)
            worker.signals.finished.connect(self._handle_matcher_results)

            # move download process to worker thread
            self.threadpool.start(worker)

    def on_input_option_change(self):
        self.__reset_widget_state()
        self._update_gene_matcher()
        self.match_genes()

    @Inputs.data_table
    def handle_input(self, data):
        self.input_data = None
        self.input_genes = None
        self.__reset_widget_state()
        self.gene_columns_model.set_domain(None)

        if data:
            self.input_data = data

            self.gene_columns_model.set_domain(self.input_data.domain)

            if self.gene_columns_model:
                self.selected_gene_col = self.gene_columns_model[0]

            self.tax_id = str(self.input_data.attributes.get(TAX_ID, ''))
            self.use_attr_names = self.input_data.attributes.get(GENE_AS_ATTRIBUTE_NAME, self.use_attr_names)

            if self.tax_id in self.organisms:
                self.selected_organism = self.organisms.index(self.tax_id)

            self.on_input_option_change()

    @staticmethod
    def get_gene_id_identifier(gene_id_strings):
        # type: (Set[str]) -> str

        if not len(gene_id_strings):
            return NCBI_ID

        regex = re.compile(r'Entrez ID \(.*?\)')
        filtered = filter(regex.search, gene_id_strings)

        return NCBI_ID + ' ({})'.format(len(set(filtered)) + 1)

    def __handle_output_data_table(self, data_table):
        """
        If 'use_attr_names' is True, genes from the input data are in columns.
        """
        # set_of_attributes = set([key for attr in data_table.domain[:] for key in attr.attributes.keys()
        #                          if key == NCBI_ID])
        # gene_id = NCBI_ID if NCBI_ID in data_table.domain or set_of_attributes else None
        if self.exclude_unmatched:
            data_table = self.__filter_unknown(data_table)

        if self.use_attr_names:
            # set_of_attributes = set([key for attr in data_table.domain[:] for key in attr.attributes.keys()
            # if key.startswith(NCBI_ID)])
            # gene_id = self.get_gene_id_identifier(set_of_attributes)
            self.gene_id_attribute = gene_id = NCBI_ID

            for gene in self.gene_matcher.genes:
                if gene.input_name in data_table.domain:

                    if gene.ncbi_id:
                        data_table.domain[gene.input_name].attributes[gene_id] = str(gene.ncbi_id)

                    if self.replace_id_with_symbol:
                        try:
                            data_table.domain[gene.input_name].name = str(gene.symbol)
                        except AttributeError:
                            pass

        else:
            set_of_variables = set([var.name for var in data_table.domain.variables + data_table.domain.metas
                                    if var.name.startswith(NCBI_ID)])

            available_rows, _ = data_table.get_column_view(self.selected_gene_col)
            self.gene_id_column = gene_id = self.get_gene_id_identifier(set_of_variables)

            temp_domain = Domain([], metas=[StringVariable(gene_id)])
            temp_data = [[str(gene.ncbi_id) if gene.ncbi_id else ''] for gene in self.gene_matcher.genes
                         if gene.input_name in available_rows]
            temp_table = Table(temp_domain, temp_data)

            # if columns differ, then concatenate.
            if NCBI_ID in data_table.domain:
                if gene_id != NCBI_ID and not np.array_equal(np.array(temp_data).ravel(),
                                                             data_table.get_column_view(NCBI_ID)[0]):

                    data_table = Table.concatenate([data_table, temp_table])
                else:
                    gene_id = NCBI_ID
            else:
                data_table = Table.concatenate([data_table, temp_table])

        return data_table, gene_id

    def __filter_unknown(self, data_table):

        known_input_genes = [gene.input_name for gene in self.gene_matcher.get_known_genes()]

        if self.use_attr_names:
            temp_domain = Domain(
                [attr for attr in data_table.domain.attributes if attr.name in known_input_genes],
                metas=data_table.domain.metas,
                class_vars=data_table.domain.class_vars
            )
            data_table = data_table.transform(temp_domain)
        else:

            # create filter from selected column for genes
            only_known = table_filter.FilterStringList(self.selected_gene_col, known_input_genes)
            # apply filter to the data
            data_table = table_filter.Values([only_known])(data_table)

        return data_table

    def commit(self):
        if self.output_data_table:
            selection = self.table_view.selectionModel().selectedRows(self.table_model.entrez_column_index)
            selected_genes = [row.data() for row in selection]

            if selected_genes:
                if self.use_attr_names:
                    selected = [column for column in self.output_data_table.domain.attributes
                                if self.gene_id_attribute in column.attributes and
                                str(column.attributes[self.gene_id_attribute]) in selected_genes]

                    domain = Domain(
                        selected, self.output_data_table.domain.class_vars,
                        self.output_data_table.domain.metas
                    )
                    new_data = self.output_data_table.from_table(domain, self.output_data_table)
                    self.Outputs.custom_data_table.send(new_data)

                else:
                    selected_rows = []
                    for row_index, row in enumerate(self.output_data_table):
                        gene_in_row = str(row[self.gene_id_column])

                        if gene_in_row in selected_genes:
                            selected_rows.append(row_index)

                    if selected_rows:
                        selected = self.output_data_table[selected_rows]
                    else:
                        selected = None
                    self.Outputs.custom_data_table.send(selected)

                self.Outputs.gene_matcher_results.send(
                    self.gene_matcher.to_data_table(selected_genes=selected_genes)
                )
            else:
                self.Outputs.gene_matcher_results.send(self.gene_matcher.to_data_table())
                self.Outputs.custom_data_table.send(self.output_data_table)

    def toggle_radio_options(self):
        self.replace_radio.setEnabled(bool(self.use_attr_names))

        if self.gene_matcher.genes:
            # enable checkbox if unknown genes are detected
            self.exclude_radio.setEnabled(len(self.gene_matcher.genes) != len(self.gene_matcher.get_known_genes()))
            self.exclude_unmatched = len(self.gene_matcher.genes) != len(self.gene_matcher.get_known_genes())

    def on_output_option_change(self):
        self.Outputs.custom_data_table.send(None)
        self.Outputs.gene_matcher_results.send(None)

        if not self.input_data:
            return

        if not self.use_attr_names and not self.gene_columns_model:
            return

        self.output_data_table = self.input_data.transform(self.input_data.domain.copy())
        self.output_data_table, gene_id = self.__handle_output_data_table(self.output_data_table.copy())

        # handle table attributes
        self.output_data_table.attributes[TAX_ID] = self.get_selected_organism()

        self.output_data_table.attributes[GENE_AS_ATTRIBUTE_NAME] = bool(self.use_attr_names)

        if not bool(self.use_attr_names):
            self.output_data_table.attributes[GENE_ID_COLUMN] = gene_id
        else:
            self.output_data_table.attributes[GENE_ID_ATTRIBUTE] = gene_id

        # print(self.output_data_table)
        self.invalidate()

    def invalidate(self):
        self.commit()




""" server update module """
import subprocess
import shutil
import sys


from urllib.request import urlopen
from server_update.common import *
from orangecontrib.bioinformatics.utils import local_cache
from unittest import TextTestRunner, makeSuite

# create_folder(update_folder)  # before calling utils.serverfiles
from orangecontrib.bioinformatics.utils.serverfiles import ServerFiles, LOCALFILES, PATH

print(PATH)
info_date_fmt = '%Y-%m-%d %H:%M:%S.%f'
http_date_fmt = '%a, %d %b %Y %H:%M:%S %Z'

# End script with exit code 10 if files on the server is up-to-date!
up_to_date = 10

sf_server = ServerFiles()
sf_local = LOCALFILES
sf_temp = 'temp'
_sync_path = 'serverfiles-bio@orange.biolab.si:/serverfiles-bio2'

# File sources
SOURCE_SERVER = 'server_file'  # files on the serverfiles-bio repository
SOURCE_USER = 'user_file'      # user defined files


def sf_last_modified(domain, filename):
    return datetime.strptime(sf_server.info(domain, filename)['datetime'], info_date_fmt)


def http_last_modified(url):
    return datetime.strptime(urlopen(url).headers.get('Last-Modified'), http_date_fmt)


class SyncHelper:

    def __init__(self, domain, test_case):
        self.domain = domain
        self.domain_path = sf_local.localpath(self.domain)
        self.test_case = test_case
        self.runner = TextTestRunner(verbosity=2)
        self.suit = makeSuite(self.test_case)
        self.results = None

    def _listfiles(self):
        """ return files that are ready to be synced to the main repository
        """
        return [fname for domain, fname in sf_local.listfiles(self.domain)]

    def _allinfo(self):
        # delete old __INFO__.txt from the server
        subprocess.call(['rsync', '-av', '--delete', '--include=__INFO__', '--exclude=*', local_cache, _sync_path])

        # create new allinfo file
        info_path = os.path.join(local_cache, '__INFO__')
        with open(info_path, 'wt') as f:
            # we must initialize ServerFiles object again because old one has __INFO__ cached
            json.dump(list(ServerFiles().allinfo().items()), f)

        subprocess.call(["rsync", info_path, _sync_path])

    def _prepare_domain_path(self):
        """ Some tests require uncompressed files to run. So we move ready(compressed) files
        to temporary folder (domain_path/temp). When/if all tests pass, we delete uncompressed files
        and move "ready to sync" files back to domain_path folder.
        """
        # temporary folder name is standardized across all update scripts
        temp_folder = os.path.join(self.domain_path, sf_temp)

        if os.path.exists(temp_folder):
            entries = os.scandir(self.domain_path)
            [os.remove(entry.path) for entry in entries if entry.is_file()]

            prepared_files = os.scandir(temp_folder)
            [shutil.move(file.path, self.domain_path) for file in prepared_files]

            try:
                os.rmdir(temp_folder)
            except OSError as e:
                # Directory not empty, copy failed?
                raise e

    def run_tests(self):
        self.results = self.runner.run(self.suit)

    def check_results(self):
        if self.results:
            return self.results.wasSuccessful()

    @staticmethod
    def remove_update_folder():
        shutil.rmtree(local_cache)

    def sync_files(self):
        """ use rsync to upload file on serverfiles-bio repo (info file included).
        """
        if not self.check_results():
            # self.remove_update_folder()
            print('Tests failed!?')
            sys.exit(1)

        self._prepare_domain_path()

        for file in self._listfiles():
            source_file = os.path.join(self.domain_path, file)
            source_info = os.path.join(self.domain_path, file + '.info')

            dest_files = os.path.join(_sync_path, self.domain)
            subprocess.call(["rsync", source_file, source_info, dest_files])
            print("Files {} and {} synced".format(file, file + '.info'))

        self._allinfo()


if __name__ == "__main__":
    # helper = SyncHelper('temp', str)
    # helper._allinfo()
    pass



from unittest import TestCase
from orangecontrib.bioinformatics.geo.dataset import GDSInfo


class GEOTest(TestCase):

    def setUp(self):
        self.info = GDSInfo()

    def test_GDS16(self):
        """ Test content of gds_info file. Check if GDS16 is correctly stored.
        """
        gds16 = self.info['GDS16']
        self.assertEqual(gds16['taxid'], '4932')
        self.assertEqual(gds16['feature_count'], 9216)
        self.assertEqual(gds16['sample_count'], 8)

    def test_size(self):
        self.assertGreater(len(self.info), 4000)



""" Bioinformatics """

ICON = "icons/Category-Bioinformatics.svg"

BACKGROUND = "light-grass"

# Location of widget help files.
WIDGET_HELP_PATH = (
    # Used for development.
    # You still need to build help pages using
    # make htmlhelp
    # inside doc folder
    ("{DEVELOP_ROOT}/doc/_build/htmlhelp/index.html", None),

    # Online documentation url, used when the local documentation is available.
    # Url should point to a page with a section Widgets. This section should
    # includes links to documentation pages of each widget. Matching is
    # performed by comparing link caption to widget name.
    ("https://orange3-bioinformatics.readthedocs.io/en/latest/", "")
)



""" Utility functions for GEO module """
import os
import io

from re import compile
from numpy import isnan, median, array, asarray
from tempfile import NamedTemporaryFile
from urllib.request import urlopen
from Orange.data import Table

from orangecontrib.bioinformatics.utils import serverfiles
from orangecontrib.bioinformatics.geo.config import DOMAIN, FTP_DIR, FTP_NCBI

unknown = float('nan')
attribute_reg = compile("![A-Za-z]*_([A-Za-z_]*) = (.*)$")


def create_table(domain, X, Y, metas):
    class_var = domain.class_var
    meta_atts = domain.metas

    if Y:
        Y = array([[class_var.to_val(row)] for row in Y], dtype=float)
    if metas:
        metas = array([[c.to_val(v) for c, v in zip(meta_atts, row)] for row in metas], dtype=object)

    return Table(domain, asarray(X), Y=Y, metas=metas)


def spots_mean(x):
    vs = [v for v in x if not isnan(v)]
    if len(vs) == 0:
        return unknown
    else:
        return sum(vs) / len(vs)


def spots_median(x):
    vs = [v for v in x if not isnan(v)]
    if len(vs) == 0:
        return unknown
    return median(vs)


def spots_min(x):
    vs = [v for v in x if not isnan(v)]
    if len(vs) == 0:
        return unknown
    else:
        return min(vs)


def spots_max(x):
    vs = [v for v in x if not isnan(v)]
    if len(vs) == 0:
        return unknown
    else:
        return max(vs)


def to_float(value):
    return float(value) if value != 'null' else unknown


def parse_attribute_line(string):
    return attribute_reg.search(string).groups()


def gds_is_cached(gds_name):
    return os.path.isfile(
        os.path.join(serverfiles.localpath(DOMAIN), gds_name + ".soft.gz"))


def sniff_size(file_obj):
    if isinstance(file_obj, io.FileIO):
        return os.fstat(file_obj.fileno()).st_size
    return None


def gds_download_url(gds_name):
    """ Return the download url for a GDS id `gdsname`.
    """
    return "ftp://{}/{}/{}.soft.gz".format(FTP_NCBI, FTP_DIR, gds_name)


def gds_ensure_downloaded(gds_name, progress=None):
    """ Ensure the GDS dataset is available locally in cache.
    """
    if gds_is_cached(gds_name):
        return
    else:
        gds_download(gds_name, progress=progress)


def gds_download(gds_name, progress=None):
    """ Download the GDS dataset into the cache.
    """
    gds_url = gds_download_url(gds_name)
    basename = gds_name + ".soft.gz"
    target_path = os.path.join(serverfiles.localpath(DOMAIN), basename)

    temp = NamedTemporaryFile(prefix=basename + "-", dir=serverfiles.localpath(DOMAIN), delete=False)
    try:
        retrieve_url(gds_url, temp, progress=progress)
    except BaseException as err:
        try:
            temp.close()
            os.remove(temp.name)
        except (OSError, IOError):
            pass
        raise err
    else:
        temp.close()
        os.replace(temp.name, target_path)


def retrieve_url(url, dest_obj, progress=None):
    """ Retrieve an `url` writing it to an open file-like `destobj`.

    Args:
        url: The source url.
        dest_obj: An file-like object opened for writing.

        progress: (int, int) -> None optional
                  An optional progress callback function. Will be called
                  periodically with `(transfered, total)` bytes count. `total`
                  can be `-1` if the total contents size cannot be
                  determined beforehand.

    Returns:

    """

    with urlopen(url, timeout=10) as stream:
        length = stream.headers.get("content-length", None)
        if length is not None:
            length = int(length)
        copyfileobj(stream, dest_obj, size=length, progress=progress)


def copyfileobj(src, dst, buffer=2 ** 15, size=None, progress=None):
    """ Like shutil.copyfileobj but with progress reporting.

    Args:
        src: Source file object
        dst: Destination file object
        buffer: Buffer size
        size: Total `src` contents size if available.

        progress: : (int, int) -> None optional
                    An optional progress callback function. Will be called
                    periodically with `(transferred, total)` bytes count. `total`
                    can be `-1` if the total contents size cannot be
                    determined beforehand.

    Returns:

    """

    count = 0
    if size is None:
        size = sniff_size(src)

    while True:
        data = src.read(buffer)
        dst.write(data)
        count += len(data)
        if progress is not None:
            progress(count, size if size is not None else -1)
        if not data:
            break

    if size is None and progress is not None:
        progress(count, count)

    return count



"""
DBGET entry
"""
from __future__ import absolute_import

import warnings
from collections import defaultdict

from . import fields
from .parser import DBGETEntryParser

__all__ = ["parser", "fields"]

# TODO: Remove the use of entry_decorate decorator
# for constructing a DBEntry subclass, make fields
# properties with __get__ method, and explicit assignment
# and meaningful docstrings


def entry_decorate(cls):
    """
    Decorate the DBEntry subclass with properties for accessing
    the fields through the 'DBField._convert' interface

    """
    reserved_names_map = {"class": "class_", "def": "def_"}

    def construct_one(name):
        def get(self):
            field = getattr(self, name, None)
            if field is not None:
                return field._convert()
            else:
                return None
        return property(get, doc=name)

    def construct_multiple(name):
        def get(self):
            field = getattr(self, name, None)
            if field is not None:
                return [f._convert() for f in field]
            else:
                return None
        return property(get, doc=name)

    for name, field in cls.FIELDS:
        name_lower = name.lower()
        if not hasattr(cls, name_lower):
            if name in cls.MULTIPLE_FIELDS:
                prop = construct_multiple(name)
            else:
                prop = construct_one(name)
            setattr(cls, reserved_names_map.get(name_lower, name_lower), prop)

    return cls


class DBEntry(object):
    """
    A DBGET entry object.
    """
    FIELDS = [("ENTRY", fields.DBEntryField)]
    MULTIPLE_FIELDS = []

    def __init__(self, text=None):
        self._sections = {}
        self.fields = []
        if text is not None:
            self.parse(text)

    @property
    def entry_key(self):
        """
        Primary entry key used for identifying the entry.
        """
        return self.entry.split(" ", 1)[0]

    def parse(self, text):
        """
        Parse `text` string containing a formated DBGET entry.
        """
        parser = DBGETEntryParser()
        gen = parser.parse_string(text)
        field_constructors = dict(self.FIELDS)

        current = None
        current_subfield = None
        entry_fields = []
        for (event, title, text) in gen:
            if event == DBGETEntryParser.SECTION_START:
                if title in field_constructors:
                    ftype = field_constructors[title]
                else:
                    ftype = fields.DBSimpleField
                current = ftype(text)
                if current.TITLE is None:
                    current.TITLE = title
            elif event == DBGETEntryParser.SECTION_END:
                entry_fields.append(current)
                current = None
            elif event == DBGETEntryParser.SUBSECTION_START:
                current_subfield = fields.DBSimpleField(text)
                current_subfield.TITLE = title
                if not isinstance(current, fields.DBFieldWithSubsections):
                    # Upgrade simple fields to FieldWithSubsection
                    new = fields.DBFieldWithSubsections(current.text)
                    new.TITLE = current.TITLE
                    current = new

            elif event == DBGETEntryParser.SUBSECTION_END:
                current.subsections.append(current_subfield)
                current_subfield = None
            elif event == DBGETEntryParser.TEXT:
                if current_subfield is not None:
                    current_subfield.text += text
                elif current is not None:
                    current.text += text
            elif event == DBGETEntryParser.ENTRY_END:
                break

        self.fields = entry_fields
        self._consolidate()

    def _consolidate(self):
        """
        Update mapping to field entries.
        """
        registered_fields = dict(self.FIELDS)
        multiple_fields = set(self.MULTIPLE_FIELDS)

        for field in self.fields:
            title = field.TITLE
            if title not in registered_fields:
                import warnings
                warnings.warn("Nonregisterd field %r in %r" % \
                              (title, type(self)))

            if title in multiple_fields:
                if not hasattr(self, title):
                    setattr(self, title, [])
                getattr(self, title).append(field)
            else:
                setattr(self, title, field)

    def __str__(self):
        return self.format()

    def format(self, section_indent=12):
        """
        Return a DBGET formated string representation.
        """
        return "".join(f.format(section_indent)
                       for f in self.fields)

    def get(self, key, default=None):
        raise NotImplementedError

        f = getattr(self, key, None)
        if f is not None:
            if key in self.MULTIPLE_FIELDS:
                return [f.text for f in f]
            else:
                return f.text
        else:
            return None



""" GeneSets utility functions """
import numpy as np


from typing import List, Tuple, NamedTuple

from orangecontrib.bioinformatics.geneset.config import GENE_SET_ATTRIBUTES
from orangecontrib.bioinformatics.utils import ensure_type
from orangecontrib.bioinformatics.utils.statistics import Hypergeometric


def filename(hierarchy, organism):  # type: (Tuple[str, str], str) -> str
    """ Obtain a filename for given hierarchy and organism.

    :param hierarchy: GeneSet hierarchy, example: ('GO', 'biological_process')
    :param organism: Taxonomy ID

    :return: Filename for given hierarchy and organism

    Example
    --------
        >>> filename(('CustomSet', 'subsets'), '6500')
        'CustomSet-subsets-6500.gmt'
    """

    return '-'.join(hierarchy + (organism if organism is not None else '',)) + '.gmt'


def filename_parse(fn):  # type: (str) -> (Tuple[Tuple[str, str], str])
    """ Returns a hierarchy and the organism from the gene set filename format.

    :param fn: GeneSets file name (.gmt)

    :return: A hierarchy and taxonomy id for given filename

    Example
    --------
        >>> filename_parse('Custom-set-6500.gmt')
        (('Custom', 'set'), '6500')
    """

    file_name = fn[:-4]  # removes .gmt extension
    parts = file_name.split('-')
    hierarchy = tuple(parts[:-1])
    org = parts[-1] if parts[-1] != '' else None
    return hierarchy, org


HYPERGEOMETRIC = Hypergeometric()

# change this when python 3.4 is not supported anymore
enrichment_result = NamedTuple('enrichment_result', [
    ('query', set),
    ('reference', set),
    ('p_value', float),
    ('enrichment_score', float)
])


class GeneSet:
    __slots__ = GENE_SET_ATTRIBUTES

    def __init__(self, gs_id=None, hierarchy=None, organism=None, name=None, genes=None, description=None, link=None):
        """ Object representing a single set of genes

        :param gs_id: Short gene set ID.
        :param hierarchy: Hierarchy should be formated as a tuple, for example ``("GO", "biological_process")``
        :param organism: Organism as a NCBI taxonomy ID.
        :param name: Gene set name.
        :param genes: A set of genes. Genes are strings.
        :param description: Gene set description.
        :param link: Link to further information about this gene set.
        """

        self.gs_id = gs_id
        self.hierarchy = hierarchy
        self.organism = organism
        self.name = name
        self.genes = genes
        self.description = description
        self.link = link

    def __hash__(self):
        return self.gs_id.__hash__() + self.name.__hash__()

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.__slots__ == other.__slots__:
                return all(getattr(self, attr) == getattr(other, attr) for attr in self.__slots__)

        return False

    def set_enrichment(self, reference, query):  # type: (List, List) -> enrichment_result
        """
        Args:
            reference:
            query:
        """

        assert len(reference) > 0
        query_mapped = self.genes.intersection(query)
        reference_mapped = self.genes.intersection(reference)

        query_p = len(query_mapped) / len(query) if query else np.nan
        ref_p = len(reference_mapped) / len(reference) if reference else np.nan
        enrichment = query_p / ref_p if ref_p else np.nan

        return enrichment_result(
            set(query_mapped), set(reference_mapped),
            HYPERGEOMETRIC.p_value(len(query_mapped), len(reference),
                                   len(reference_mapped), len(query)),
            enrichment)

    def gmt_description(self):
        """ Represent GeneSet as line in GMT file format

        :return: Comma-separated GeneSet attributes.
        """

        # this is just so that we preserve order from GENE_SET_ATTRIBUTES for easy access
        empty_field = '_'
        hierarchy = '-'.join((hier for hier in self.hierarchy)) if self.hierarchy else empty_field

        return ','.join((self.gs_id,
                         hierarchy,
                         self.organism,
                         self.name if self.name else empty_field,
                         empty_field,
                         self.description if self.description else empty_field,
                         self.link if self.link else empty_field))


class GeneSets(set):
    """ A collection of gene sets: contains :obj:`GeneSet` objects.
    """
    
    def __init__(self, sets=None):
        # type: (List[GeneSet]) -> None
        super().__init__()

        if sets:
            self.update(ensure_type(sets, list))

    def update(self, sets):
        # type: (List[GeneSet]) -> None

        for g_set in sets:
            self.add(ensure_type(g_set, GeneSet))

    def common_org(self):
        """ Return a common organism. """
        if len(self) == 0:
            raise GeneSetException("Empty gene sets.")

        organisms = set(g_set.organism for g_set in self)
        if not len(organisms) == 1:
            raise GeneSetException("multiple organisms in a set " + str(organisms))
        else:
            for org in organisms:
                return org

    def hierarchies(self):
        """ Return all hierarchies. """
        if len(self) == 0:
            raise GeneSetException("Empty gene sets.")

        return set(g_set.hierarchy for g_set in self)

    def common_hierarchy(self):
        """ Return a common hierarchy. """
        hierarchies = self.hierarchies()

        if not len(hierarchies) == 1:
            raise GeneSetException('multiple hierarchies in a set', str(hierarchies))
        else:
            for org in hierarchies:
                return org

    def delete_sets_by_hierarchy(self, hier):
        selected_sets = self.map_hierarchy_to_sets().get(hier, None)
        if selected_sets:
            [self.remove(gene_set) for gene_set in selected_sets]

    def map_hierarchy_to_sets(self):
        try:
            split_by_hier = {hier: GeneSets() for hier in self.hierarchies()}
            [split_by_hier[gs.hierarchy].update([gs]) for gs in self]
            return split_by_hier

        except GeneSetException:
            return {}

    def split_by_hierarchy(self):
        """ Split gene sets by hierarchies. Return a list of :class:`GeneSets` objects. """

        try:
            split_by_hier = {hier: GeneSets() for hier in self.hierarchies()}
            [split_by_hier[gs.hierarchy].update([gs]) for gs in self]
            return list(split_by_hier.values())

        except GeneSetException:
            return []

    def genes(self):
        """
        Returns:
            All genes from GeneSets
        """
        genes = set()
        [genes.update(gene_set.genes) for gene_set in self]
        return genes

    def to_gmt_file_format(self, file_path):  # type: (str) -> None
        """ The GMT file format is a tab delimited file format that describes gene sets.

        In the GMT format, each row represents a gene set.
        Columns: gs_id    gmt_description    Gene    Gene    Gene    ...
        gmt_description: 'gs_id','hierarchy','organism','name','genes','description','link'

        :param file_path: Path to where file will be created

        """

        with open(file_path, 'a') as gmt_file:
            for gene_set in self:
                genes = map(str, gene_set.genes)
                line = '\t'.join([gene_set.gs_id, gene_set.gmt_description()] + list(genes))
                gmt_file.write(line + '\n')

    @staticmethod
    def from_gmt_file_format(file_path):  # type: (str) -> GeneSets
        """ Load GeneSets object from GMT file.

        :param file_path: path to a file on local disk
        :rtype: :obj:`GeneSets`
        """
        index = {label: index for index, label in enumerate(GENE_SET_ATTRIBUTES)}

        with open(file_path, 'r', encoding='utf-8') as gmt_file:
            gene_sets = []

            for line in gmt_file:
                columns = [column.strip() for column in line.split('\t')]
                gs_info = columns[1].split(',')
                hierarchy = tuple(gs_info[index['hierarchy']].split('-'))
                genes = set([str(gene) for gene in columns[2:]])

                gene_set = GeneSet(gs_id=columns[0],
                                   genes=genes,
                                   hierarchy=hierarchy,
                                   name=gs_info[index['name']],
                                   organism=gs_info[index['organism']],
                                   description=gs_info[index['description']],
                                   link=gs_info[index['link']])

                gene_sets.append(gene_set)

            return GeneSets(gene_sets)


class NoGeneSetsException(Exception):
    """ Raised when provided taxonomy is not in orangecontrib.bio.ncbi.taxonomy.common_taxids """


class GeneSetException(Exception):
    pass


if __name__ == '__main__':
    pass



from orangecontrib.bioinformatics.ncbi.gene import GeneMatcher, GENE_INFO_TAGS

# specify input
organism = 9606
genes_symbols_to_match = ['HB1', 'BCKDHB', 'TWIST1']

# initialize gene matcher object
gene_matcher = GeneMatcher(organism)
gene_matcher.genes = genes_symbols_to_match

# run matching process
gene_matcher.run_matcher()

# inspect results
for gene in gene_matcher.genes:
    print("\ninput name: " + gene.input_name,
          "\nid from ncbi: ", gene.ncbi_id,
          "\nmatch type: ", gene.type_of_match
          )
    if gene.ncbi_id is None and gene.possible_hits:
        print('possible_hits: ', [hit.ncbi_id for hit in gene.possible_hits])



DOMAIN = 'gene_info'
FILENAME = 'gene_info.sqlite'
TITLE = 'NCBI Gene database'
TAGS = ['NCBI', 'Gene names', 'Gene Ids', 'Description', 'Essential']
FTP_FILENAME = 'gene_info.gz'
FTP_URL = 'ftp://ftp.ncbi.nlm.nih.gov/gene/DATA/gene_info.gz'


# NCBI_INFO - order is important
GENE_INFO_TAGS = ['tax_id',
                  'gene_id',
                  'symbol',
                  'synonyms',
                  'db_refs',
                  'description',
                  'locus_tag',
                  'chromosome',
                  'map_location',
                  'type_of_gene',
                  'symbol_from_nomenclature_authority',
                  'full_name_from_nomenclature_authority',
                  'nomenclature_status',
                  'other_designations',
                  'modification_date']

GENE_MATCHER_TAGS = ['input_name', 'type_of_match', 'ncbi_id', '_possible_hits']

GENE_MATCHER_HEADER = [
    ['Input ID', 'Entrez ID', 'Name', 'Description', 'Synonyms', 'Other IDs'],
    ['input_name', 'ncbi_id', 'symbol', 'description', 'synonyms', 'db_refs']
]

# PICKLED GENE MAPPER
MATCHER_FILENAME = 'gene_mapper.{}'
MATCHER_TITLE = 'Gene mapper'
MATCHER_TAGS = ['source id', 'ncbi id', 'symbols', 'synonyms', 'genes', 'mapping']

MAP_TAX_ID = 'tax_id'
MAP_GENE_ID = 'gene_id'
MAP_SOURCES = 'sources'
MAP_SYMBOL = 'symbol'
MAP_SYNONYMS = 'synonyms'
MAP_LOCUS = 'locus_tag'
MAP_NOMENCLATURE = 'symbol_from_nomenclature_authority'

# Pretty strings
NCBI_ID = 'Entrez ID'
ENSEMBl_ID = 'Ensembl ID'
GENE_SYMBOL = 'Symbol'
GENE_SYNONYMS = 'Synonyms'

# OWGeneInfo settings
GENE_INFO_HEADER_LABELS = ['NCBI ID', 'Ensembl ID', 'Symbol', 'Locus Tag', 'Chromosome', 'Description', 'Synonyms', 'Nomenclature']
NCBI_DETAIL_LINK = 'http://www.ncbi.nlm.nih.gov/sites/entrez?Db=gene&Cmd=ShowDetailView&TermToSearch={}'
ENSEMBL_DETAIL_LINK = 'http://www.ensembl.org/id/{}'



""" Gene Expression Omnibus datasets widget """
import glob
import os
import sys
import numpy
import urllib.request as urlrequest


from collections import defaultdict
from functools import partial, lru_cache

from AnyQt.QtWidgets import (
    QLineEdit, QSplitter, QTreeView, QTreeWidget, QTreeWidgetItem,
)
from AnyQt.QtCore import (
    Qt, QThread, QCoreApplication, QSortFilterProxyModel, QItemSelectionModel, Slot
)
from AnyQt.QtGui import QStandardItemModel, QStandardItem

from Orange.data import Table, Domain, StringVariable
from Orange.widgets import gui
from Orange.widgets.widget import OWWidget
from Orange.widgets.settings import Setting
from Orange.widgets.utils.concurrent import ThreadExecutor, Task, methodinvoke

from orangecontrib.bioinformatics.utils import serverfiles
from orangecontrib.bioinformatics.geo.utils import gds_ensure_downloaded
from orangecontrib.bioinformatics.geo.dataset import GDS, GDSInfo, DOMAIN
from orangecontrib.bioinformatics.widgets.utils.gui import TokenListCompleter
from orangecontrib.bioinformatics.widgets.utils.data import (
    GENE_AS_ATTRIBUTE_NAME, TAX_ID, GENE_ID_COLUMN, GENE_ID_ATTRIBUTE
)
from orangecontrib.bioinformatics.ncbi.gene import NCBI_ID


TextFilterRole = next(gui.OrangeUserRole)


class MySortFilterProxyModel(QSortFilterProxyModel):
    def __init__(self, parent=None):
        QSortFilterProxyModel.__init__(self, parent)
        self._filter_strings = []
        self._cache = {}
        self._cache_fixed = {}
        self._cache_prefix = {}
        self._row_text = {}

        # Create a cached version of _filteredRows
        self._filteredRows = lru_cache(100)(self._filteredRows)

    def setSourceModel(self, model):
        """ Set the source model for the filter.
        """
        self._filter_strings = []
        self._cache = {}
        self._cache_fixed = {}
        self._cache_prefix = {}
        self._row_text = {}
        QSortFilterProxyModel.setSourceModel(self, model)

    def addFilterFixedString(self, string, invalidate=True):
        """ Add `string` filter to the list of filters. If invalidate is
        True the filter cache will be recomputed.
        """
        self._filter_strings.append(string)
        all_rows = range(self.sourceModel().rowCount())
        row_text = [self.rowFilterText(row) for row in all_rows]
        self._cache[string] = [string in text for text in row_text]
        if invalidate:
            self.updateCached()
            self.invalidateFilter()

    def removeFilterFixedString(self, index=-1, invalidate=True):
        """ Remove the `index`-th filter string. If invalidate is True the
        filter cache will be recomputed.
        """
        string = self._filter_strings.pop(index)
        del self._cache[string]
        if invalidate:
            self.updateCached()
            self.invalidate()

    def setFilterFixedStrings(self, strings):
        """Set a list of string to be the new filters.
        """
        to_remove = set(self._filter_strings) - set(strings)
        to_add = set(strings) - set(self._filter_strings)
        for str in to_remove:
            self.removeFilterFixedString(
                self._filter_strings.index(str),
                invalidate=False)

        for str in to_add:
            self.addFilterFixedString(str, invalidate=False)
        self.updateCached()
        self.invalidate()

    def _filteredRows(self, filter_strings):
        """Return a dictionary mapping row indexes to True False values.

        .. note:: This helper function is wrapped in the __init__ method.

        """
        all_rows = range(self.sourceModel().rowCount())
        cache = self._cache
        return dict([(row, all([cache[str][row] for str in filter_strings]))
                     for row in all_rows])

    def updateCached(self):
        """Update the combined filter cache.
        """
        self._cache_fixed = self._filteredRows(
            tuple(sorted(self._filter_strings)))

    def setFilterFixedString(self, string):
        """Should this raise an error? It is not being used.
        """
        QSortFilterProxyModel.setFilterFixedString(self, string)

    def rowFilterText(self, row):
        """Return text for `row` to filter on.
        """
        f_role = self.filterRole()
        f_column = self.filterKeyColumn()
        s_model = self.sourceModel()
        data = s_model.data(s_model.index(row, f_column), f_role)
        return str(data)

    def filterAcceptsRow(self, row, parent):
        return self._cache_fixed.get(row, True)

    def lessThan(self, left, right):
        # TODO: Remove fixed column handling
        if left.column() == 1 and right.column():
            left_gds = str(left.data(Qt.DisplayRole))
            right_gds = str(right.data(Qt.DisplayRole))
            left_gds = left_gds.lstrip("GDS")
            right_gds = right_gds.lstrip("GDS")
            try:
                return int(left_gds) < int(right_gds)
            except ValueError:
                pass
        return QSortFilterProxyModel.lessThan(self, left, right)


def childiter(item):
    """ Iterate over the children of an QTreeWidgetItem instance.
    """
    for i in range(item.childCount()):
        yield item.child(i)


class OWGEODatasets(OWWidget):
    name = "GEO Data Sets"
    description = "Access to Gene Expression Omnibus data sets."
    icon = "icons/OWGEODatasets.svg"
    priority = 2

    inputs = []
    outputs = [("Expression Data", Table)]

    settingsList = ["outputRows", "mergeSpots", "gdsSelectionStates",
                    "splitterSettings", "currentGds", "autoCommit",
                    "datasetNames"]

    outputRows = Setting(True)
    mergeSpots = Setting(True)
    gdsSelectionStates = Setting({})
    currentGds = Setting(None)
    datasetNames = Setting({})
    splitterSettings = Setting(
        (b'\x00\x00\x00\xff\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x01\xea\x00\x00\x00\xd7\x01\x00\x00\x00\x07\x01\x00\x00\x00\x02',
         b'\x00\x00\x00\xff\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x01\xb5\x00\x00\x02\x10\x01\x00\x00\x00\x07\x01\x00\x00\x00\x01')
    )

    autoCommit = Setting(False)

    def __init__(self, parent=None, signalManager=None, name=" GEO Data Sets"):
        OWWidget.__init__(self, parent, signalManager, name)

        self.selectionChanged = False
        self.filterString = ""
        self.datasetName = ""

        ## GUI
        box = gui.widgetBox(self.controlArea, "Info", addSpace=True)
        self.infoBox = gui.widgetLabel(box, "Initializing\n\n")

        box = gui.widgetBox(self.controlArea, "Output", addSpace=True)
        gui.radioButtonsInBox(box, self, "outputRows",
                              ["Genes in rows", "Samples in rows"], "Rows",
                              callback=self.commitIf)

        gui.checkBox(box, self, "mergeSpots", "Merge spots of same gene",
                     callback=self.commitIf)

        gui.separator(box)
        self.nameEdit = gui.lineEdit(
            box, self, "datasetName", "Data set name",
            tooltip="Override the default output data set name",
            callback=self.onNameEdited
        )
        self.nameEdit.setPlaceholderText("")

        if sys.version_info < (3, ):
            box = gui.widgetBox(self.controlArea, "Commit", addSpace=True)
            self.commitButton = gui.button(
                box, self, "Commit", callback=self.commit)
            cb = gui.checkBox(box, self, "autoCommit", "Commit on any change")
            gui.setStopper(self, self.commitButton, cb, "selectionChanged",
                           self.commit)
        else:
            gui.auto_commit(self.controlArea, self, "autoCommit", "Commit",
                            box="Commit")
            self.commitIf = self.commit

        gui.rubber(self.controlArea)

        gui.widgetLabel(self.mainArea, "Filter")
        self.filterLineEdit = QLineEdit(
            textChanged=self.filter
        )
        self.completer = TokenListCompleter(
            self, caseSensitivity=Qt.CaseInsensitive
        )
        self.filterLineEdit.setCompleter(self.completer)

        self.mainArea.layout().addWidget(self.filterLineEdit)

        splitter = QSplitter(Qt.Vertical, self.mainArea)
        self.mainArea.layout().addWidget(splitter)
        self.treeWidget = QTreeView(splitter)

        self.treeWidget.setSelectionMode(QTreeView.SingleSelection)
        self.treeWidget.setRootIsDecorated(False)
        self.treeWidget.setSortingEnabled(True)
        self.treeWidget.setAlternatingRowColors(True)
        self.treeWidget.setUniformRowHeights(True)
        self.treeWidget.setEditTriggers(QTreeView.NoEditTriggers)

        linkdelegate = gui.LinkStyledItemDelegate(self.treeWidget)
        self.treeWidget.setItemDelegateForColumn(1, linkdelegate)
        self.treeWidget.setItemDelegateForColumn(8, linkdelegate)
        self.treeWidget.setItemDelegateForColumn(
            0, gui.IndicatorItemDelegate(self.treeWidget,
                                         role=Qt.DisplayRole))

        proxyModel = MySortFilterProxyModel(self.treeWidget)
        self.treeWidget.setModel(proxyModel)
        self.treeWidget.selectionModel().selectionChanged.connect(
            self.updateSelection
        )
        self.treeWidget.viewport().setMouseTracking(True)

        splitterH = QSplitter(Qt.Horizontal, splitter)

        box = gui.widgetBox(splitterH, "Description")
        self.infoGDS = gui.widgetLabel(box, "")
        self.infoGDS.setWordWrap(True)
        gui.rubber(box)

        box = gui.widgetBox(splitterH, "Sample Annotations")
        self.annotationsTree = QTreeWidget(box)
        self.annotationsTree.setHeaderLabels(
            ["Type (Sample annotations)", "Sample count"]
        )
        self.annotationsTree.setRootIsDecorated(True)
        box.layout().addWidget(self.annotationsTree)
        self.annotationsTree.itemChanged.connect(
            self.annotationSelectionChanged
        )
        self._annotationsUpdating = False
        self.splitters = splitter, splitterH

        for sp, setting in zip(self.splitters, self.splitterSettings):
            sp.splitterMoved.connect(self.splitterMoved)
            sp.restoreState(setting)

        self.searchKeys = ["dataset_id", "title", "platform_organism",
                           "description"]

        self.gds = []
        self.gds_info = None

        self.resize(1000, 600)

        self.setBlocking(True)
        self.setEnabled(False)
        self.progressBarInit()

        self._executor = ThreadExecutor()

        func = partial(get_gds_model,
                       methodinvoke(self, "_setProgress", (float,)))
        self._inittask = Task(function=func)
        self._inittask.finished.connect(self._initializemodel)
        self._executor.submit(self._inittask)

        self._datatask = None

    @Slot(float)
    def _setProgress(self, value):
        self.progressBarValue = value

    def _initializemodel(self):
        assert self.thread() is QThread.currentThread()
        model, self.gds_info, self.gds = self._inittask.result()
        model.setParent(self)

        proxy = self.treeWidget.model()
        proxy.setFilterKeyColumn(0)
        proxy.setFilterRole(TextFilterRole)
        proxy.setFilterCaseSensitivity(False)
        proxy.setFilterFixedString(self.filterString)

        proxy.setSourceModel(model)
        proxy.sort(0, Qt.DescendingOrder)

        self.progressBarFinished()
        self.setBlocking(False)
        self.setEnabled(True)

        filter_items = " ".join(
            gds[key] for gds in self.gds for key in self.searchKeys
        )
        tr_chars = ",.:;!?(){}[]_-+\\|/%#@$^&*<>~`"
        tr_table = str.maketrans(tr_chars, " " * len(tr_chars))
        filter_items = filter_items.translate(tr_table)

        filter_items = sorted(set(filter_items.split(" ")))
        filter_items = [item for item in filter_items if len(item) > 3]

        self.completer.setTokenList(filter_items)

        if self.currentGds:
            current_id = self.currentGds["dataset_id"]
            gdss = [(i, proxy.data(proxy.index(i, 1), Qt.DisplayRole))
                    for i in range(proxy.rowCount())]
            current = [i for i, data in gdss if data and data == current_id]
            if current:
                current_index = proxy.index(current[0], 0)
                self.treeWidget.selectionModel().select(
                    current_index,
                    QItemSelectionModel.Select | QItemSelectionModel.Rows
                )
                self.treeWidget.scrollTo(
                    current_index, QTreeView.PositionAtCenter)

        for i in range(8):
            self.treeWidget.resizeColumnToContents(i)

        self.treeWidget.setColumnWidth(
            1, min(self.treeWidget.columnWidth(1), 300))
        self.treeWidget.setColumnWidth(
            2, min(self.treeWidget.columnWidth(2), 200))

        self.updateInfo()

    def updateInfo(self):
        gds_info = self.gds_info
        text = ("%i datasets\n%i datasets cached\n" %
                (len(gds_info),
                 len(glob.glob(serverfiles.localpath("GEO") + "/GDS*"))))
        filtered = self.treeWidget.model().rowCount()
        if len(self.gds) != filtered:
            text += ("%i after filtering") % filtered
        self.infoBox.setText(text)

    def updateSelection(self, *args):
        current = self.treeWidget.selectedIndexes()
        mapToSource = self.treeWidget.model().mapToSource
        current = [mapToSource(index).row() for index in current]
        if current:
            self.currentGds = self.gds[current[0]]
            self.setAnnotations(self.currentGds)
            self.infoGDS.setText(self.currentGds.get("description", ""))
            self.nameEdit.setPlaceholderText(self.currentGds["title"])
            self.datasetName = \
                self.datasetNames.get(self.currentGds["dataset_id"], "")
        else:
            self.currentGds = None
            self.nameEdit.setPlaceholderText("")
            self.datasetName = ""

        self.commitIf()

    def setAnnotations(self, gds):
        self._annotationsUpdating = True
        self.annotationsTree.clear()

        annotations = defaultdict(set)
        subsetscount = {}
        for desc in gds["subsets"]:
            annotations[desc["type"]].add(desc["description"])
            subsetscount[desc["description"]] = str(len(desc["sample_id"]))

        for type, subsets in annotations.items():
            key = (gds["dataset_id"], type)
            subsetItem = QTreeWidgetItem(self.annotationsTree, [type])
            subsetItem.setFlags(subsetItem.flags() | Qt.ItemIsUserCheckable |
                                Qt.ItemIsTristate)
            subsetItem.setCheckState(
                0, self.gdsSelectionStates.get(key, Qt.Checked)
            )
            subsetItem.key = key
            for subset in subsets:
                key = (gds["dataset_id"], type, subset)
                item = QTreeWidgetItem(
                    subsetItem, [subset, subsetscount.get(subset, "")]
                )
                item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
                item.setCheckState(
                    0, self.gdsSelectionStates.get(key, Qt.Checked)
                )
                item.key = key
        self._annotationsUpdating = False
        self.annotationsTree.expandAll()
        for i in range(self.annotationsTree.columnCount()):
            self.annotationsTree.resizeColumnToContents(i)

    def annotationSelectionChanged(self, item, column):
        if self._annotationsUpdating:
            return
        for i in range(self.annotationsTree.topLevelItemCount()):
            item = self.annotationsTree.topLevelItem(i)
            self.gdsSelectionStates[item.key] = item.checkState(0)
            for j in range(item.childCount()):
                child = item.child(j)
                self.gdsSelectionStates[child.key] = child.checkState(0)

    def filter(self):
        filter_string = self.filterLineEdit.text()
        proxyModel = self.treeWidget.model()
        if proxyModel:
            strings = filter_string.lower().strip().split()
            proxyModel.setFilterFixedStrings(strings)
            self.updateInfo()

    def selectedSamples(self):
        """
        Return the currently selected sample annotations.

        The return value is a list of selected (sample type, sample value)
        tuples.

        .. note:: if some Sample annotation type has no selected values.
                  this method will return all values for it.

        """
        samples = []
        unused_types = []
        used_types = []
        for stype in childiter(self.annotationsTree.invisibleRootItem()):
            selected_values = []
            all_values = []
            for sval in childiter(stype):
                value = (str(stype.text(0)), str(sval.text(0)))
                if self.gdsSelectionStates.get(sval.key, True):
                    selected_values.append(value)
                all_values.append(value)
            if selected_values:
                samples.extend(selected_values)
                used_types.append(str(stype.text(0)))
            else:
                # If no sample of sample type is selected we don't filter
                # on it.
                samples.extend(all_values)
                unused_types.append(str(stype.text(0)))

        return samples, used_types

    def commitIf(self):
        if self.autoCommit:
            self.commit()
        else:
            self.selectionChanged = True

    @Slot(int, int)
    def progressCompleted(self, value, total):
        if total > 0:
            self.progressBarSet(100. * value / total, processEvents=False)
        else:
            pass
            # TODO: report 'indeterminate progress'

    def commit(self):
        if self.currentGds:
            self.error(0)
            sample_type = None
            self.progressBarInit(processEvents=None)

            _, groups = self.selectedSamples()
            if len(groups) == 1 and self.outputRows:
                sample_type = groups[0]

            self.setEnabled(False)
            self.setBlocking(True)

            progress = methodinvoke(self, "progressCompleted", (int, int))

            def get_data(gds_id, report_genes, transpose, sample_type, title):
                gds_ensure_downloaded(gds_id, progress)
                gds = GDS(gds_id)
                data = gds.get_data(
                    report_genes=report_genes, transpose=transpose,
                    sample_type=sample_type
                )
                data.name = title
                return data

            get_data = partial(
                get_data, self.currentGds["dataset_id"],
                report_genes=self.mergeSpots,
                transpose=self.outputRows,
                sample_type=sample_type,
                title=self.datasetName or self.currentGds["title"]
            )
            self._datatask = Task(function=get_data)
            self._datatask.finished.connect(self._on_dataready)
            self._executor.submit(self._datatask)

    def _on_dataready(self):
        self.setEnabled(True)
        self.setBlocking(False)
        self.progressBarFinished(processEvents=False)

        try:
            data = self._datatask.result()
        except urlrequest.URLError as error:
            self.error(0, ("Error while connecting to the NCBI ftp server! "
                           "'%s'" % error))
            sys.excepthook(type(error), error, getattr(error, "__traceback__"))
            return
        finally:
            self._datatask = None

        data_name = data.name
        samples, _ = self.selectedSamples()

        self.warning(0)
        message = None
        from orangecontrib.bioinformatics.ncbi.gene import GeneMatcher

        gene_matcher = GeneMatcher(self.currentGds.get('taxid', ''))

        if self.outputRows:
            def samplesinst(ex):
                out = []
                for meta in data.domain.metas:
                    out.append((meta.name, ex[meta].value))

                if data.domain.class_var.name != 'class':
                    out.append((data.domain.class_var.name,
                                ex[data.domain.class_var].value))

                return out
            samples = set(samples)
            mask = [samples.issuperset(samplesinst(ex)) for ex in data]
            data = data[numpy.array(mask, dtype=bool)]
            gene_matcher.match_table_attributes(data)
            if len(data) == 0:
                message = "No samples with selected sample annotations."
        else:
            samples = set(samples)
            domain = Domain(
                [attr for attr in data.domain.attributes
                 if samples.issuperset(attr.attributes.items())],
                data.domain.class_var,
                data.domain.metas
            )
#             domain.addmetas(data.domain.getmetas())

            if len(domain.attributes) == 0:
                message = "No samples with selected sample annotations."
            stypes = set(s[0] for s in samples)
            for attr in domain.attributes:
                attr.attributes = dict(
                    (key, value) for key, value in attr.attributes.items()
                    if key in stypes
                )

            data = Table(domain, data)

            if 'gene' in data.domain:
                gene_column = data.domain['gene']
                gene_names = data.get_column_view(gene_column)[0]
                gene_matcher.genes = gene_names
                gene_matcher.run_matcher()

                domain_ids = Domain([], metas=[StringVariable(NCBI_ID)])
                data_ids = [[str(gene.ncbi_id) if gene.ncbi_id else '?'] for gene in gene_matcher.genes]
                table_ids = Table(domain_ids, data_ids)

                data = Table.concatenate([data, table_ids])

        if message is not None:
            self.warning(0, message)

        data.attributes[TAX_ID] = self.currentGds.get('taxid', '')
        data.attributes[GENE_AS_ATTRIBUTE_NAME] = bool(self.outputRows)

        if not bool(self.outputRows):
            data.attributes[GENE_ID_COLUMN] = NCBI_ID
        else:
            data.attributes[GENE_ID_ATTRIBUTE] = NCBI_ID

        data.name = data_name
        self.send("Expression Data", data)

        model = self.treeWidget.model().sourceModel()
        row = self.gds.index(self.currentGds)

        model.setData(model.index(row, 0),  " ", Qt.DisplayRole)

        self.updateInfo()
        self.selectionChanged = False


    def splitterMoved(self, *args):
        self.splitterSettings = [bytes(sp.saveState()) for sp in self.splitters]

    def send_report(self):
        self.report_items("GEO Dataset",
                          [("ID", self.currentGds['dataset_id']), ("Title", self.currentGds['title']),
                           ("Organism", self.currentGds['sample_organism'])])
        self.report_items("Data", [("Samples", self.currentGds['sample_count']),
                                   ("Features", self.currentGds['feature_count']),
                                   ("Genes", self.currentGds['gene_count'])])
        self.report_name("Sample annotations")
        subsets = defaultdict(list)
        for subset in self.currentGds['subsets']:
            subsets[subset['type']].append((subset['description'], len(subset['sample_id'])))
        self.report_html += "<ul>"
        for type in subsets:
            self.report_html += "<b>" + type + ":</b></br>"
            for desc, count in subsets[type]:
                self.report_html += 9 * "&nbsp" + "<b>{}:</b> {}</br>".format(desc, count)
        self.report_html += "</ul>"

    def onDeleteWidget(self):
        if self._inittask:
            self._inittask.future().cancel()
            self._inittask.finished.disconnect(self._initializemodel)
        if self._datatask:
            self._datatask.future().cancel()
            self._datatask.finished.disconnect(self._on_dataready)
        self._executor.shutdown(wait=False)

        super(OWGEODatasets, self).onDeleteWidget()

    def onNameEdited(self):
        if self.currentGds:
            gds_id = self.currentGds["dataset_id"]
            self.datasetNames[gds_id] = self.nameEdit.text()
            self.commitIf()


def get_gds_model(progress=lambda val: None):
    """
    Initialize and return a GDS datasets model.

    :param progress: A progress callback.
    :rval tuple:
        A tuple of (QStandardItemModel, GDSInfo, [GDS])

    .. note::
        The returned QStandardItemModel's thread affinity is set to
        the GUI thread.

    """
    progress(1)
    info = GDSInfo()
    search_keys = ["dataset_id", "title", "platform_organism", "description"]
    cache_dir = serverfiles.localpath(DOMAIN)
    gds_link = "http://www.ncbi.nlm.nih.gov/sites/GDSbrowser?acc={0}"
    pm_link = "http://www.ncbi.nlm.nih.gov/pubmed/{0}"
    gds_list = []

    def is_cached(gds):
        return os.path.exists(os.path.join(cache_dir, gds["dataset_id"]) +
                              ".soft.gz")

    def item(displayvalue, item_values={}):
        item = QStandardItem()
        item.setData(displayvalue, Qt.DisplayRole)
        for role, value in item_values.items():
            item.setData(value, role)
        return item

    def gds_to_row(gds):
        #: Text for easier full search.
        search_text = " | ".join([gds.get(key, "").lower()
                                  for key in search_keys])
        row = [
            item(" " if is_cached(gds) else "",
                 {TextFilterRole: search_text}),
            item(gds["dataset_id"],
                 {gui.LinkRole: gds_link.format(gds["dataset_id"])}),
            item(gds["title"]),
            item(gds["platform_organism"]),
            item(len(gds["samples"])),
            item(gds["feature_count"]),
            item(gds["gene_count"]),
            item(len(gds["subsets"])),
            item(gds.get("pubmed_id", ""),
                 {gui.LinkRole: pm_link.format(gds["pubmed_id"])
                            if gds.get("pubmed_id")
                            else None})
        ]
        return row

    model = QStandardItemModel()
    model.setHorizontalHeaderLabels(
        ["", "ID", "Title", "Organism", "Samples", "Features",
         "Genes", "Subsets", "PubMedID"]
    )
    progress(20)
    for gds in info.values():
        model.appendRow(gds_to_row(gds))

        gds_list.append(gds)

    progress(50)

    if QThread.currentThread() is not QCoreApplication.instance().thread():
        model.moveToThread(QCoreApplication.instance().thread())
    return model, info, gds_list


if __name__ == "__main__":
    def main_test():
        from AnyQt.QtWidgets import QApplication
        app = QApplication([])
        w = OWGEODatasets()
        w.show()
        w.raise_()
        r = app.exec_()
        w.saveSettings()
        return r

    sys.exit(main_test())



"""ServerFiles"""
import os
import serverfiles


from orangecontrib.bioinformatics.utils import local_cache

server_url = "http://orange.biolab.si/serverfiles-bio2/"


class ServerFiles(serverfiles.ServerFiles):

    def __init__(self, server=server_url):
        serverfiles.ServerFiles.__init__(self, server)


PATH = os.path.join(local_cache, 'serverfiles-bio')
LOCALFILES = serverfiles.LocalFiles(PATH, serverfiles=ServerFiles())


def localpath(*args, **kwargs):
    return LOCALFILES.localpath(*args, **kwargs)

    
def listfiles(*args):
    files = serverfiles.LocalFiles(PATH, serverfiles=ServerFiles())
    return files.listfiles(*args)


def localpath_download(*path, **kwargs):
    files = serverfiles.LocalFiles(PATH, serverfiles=ServerFiles())
    return files.localpath_download(*path, **kwargs)


def download(*path, **kwargs):
    files = serverfiles.LocalFiles(PATH, serverfiles=ServerFiles())
    return files.download(*path, **kwargs)


def allinfo(*args, **kwargs):
    files = serverfiles.LocalFiles(PATH, serverfiles=ServerFiles())
    return files.allinfo(*args, **kwargs)


def info(*args, **kwargs):
    files = serverfiles.LocalFiles(PATH, serverfiles=ServerFiles())
    return files.info(*args, **kwargs)


def need_update(*path):
    files = serverfiles.LocalFiles(PATH, serverfiles=ServerFiles())
    return files.needs_update(*path)


def update(*path, **kwargs):
    files = serverfiles.LocalFiles(PATH, serverfiles=ServerFiles())
    return files.update(*path, **kwargs)


def sizeformat(size):
    return serverfiles.sizeformat(size)



""" Multithreading widgets with QThreadPool """

from AnyQt.QtCore import (
    QObject, pyqtSignal, QRunnable, pyqtSlot
)


class Signals(QObject):
    progress = pyqtSignal()
    partial_result = pyqtSignal(object)
    result = pyqtSignal(object)
    finished = pyqtSignal()
    error = pyqtSignal(Exception)


class Worker(QRunnable):
    """ Worker runnable
    """

    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()
        # Store constructor arguments (re-used for processing)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = Signals()

        is_callback = kwargs.get("progress_callback", None)
        is_partial = kwargs.get("partial_result", None)

        if is_callback:
            self.kwargs['progress_callback'] = self.signals.progress

        if is_partial:
            self.kwargs['partial_result'] = self.signals.partial_result

    @pyqtSlot()
    def run(self):
        try:
            result = self.fn(*self.args, **self.kwargs)
        except Exception as e:
            self.signals.error.emit(e)
        else:
            self.signals.result.emit(result)
        finally:
            self.signals.finished.emit()



from orangecontrib.bioinformatics import go

ontology = go.Ontology()
annotations = go.Annotations("4932", ontology=ontology)

# keys are symbol names, values are Entrez IDs
genes_ids = {'Yta7p': '853186', 'RPN2': '854735', 'RPT2': '851557'}
res = annotations.get_enriched_terms(genes_ids.values())

print(res)
print("Enriched terms:")
for go_id, (genes, p_value, ref) in res.items():
    if p_value < 0.05:
        print(ontology[go_id].name + " with p-value: %.4f " % p_value + ", ".join(genes))

# And again for slims
annotations.ontology.set_slims_subset('goslim_yeast')

res = annotations.get_enriched_terms(genes_ids.values(), slims_only=True)
print("\n\nEnriched slim terms:")
for go_id, (genes, p_value, _) in res.items():
    if p_value < 0.2:
        print(ontology[go_id].name + " with p-value: %.4f " % p_value + ", ".join(genes))



""" Qt component for Gene scoring method selection """
from collections import namedtuple
from typing import Union

from AnyQt.QtWidgets import (
    QWidget
)

from Orange.widgets.gui import (
    radioButtons, QGroupBox, comboBox
)


from orangecontrib.bioinformatics.utils.statistics import score_t_test, score_mann_whitney, score_hypergeometric_test, \
    ALTERNATIVES, ALT_LESS, ALT_TWO, ALT_GREATER


#: Selection types
LowTail, HighTail, TwoTail = ALT_LESS, ALT_TWO, ALT_GREATER
#: Test type - i.e a two sample (t-test, ...) or multi-sample (ANOVA) test
TwoSampleTest, VarSampleTest = 1, 2

gene_scoring_method = namedtuple('scoring_method', 'name, score_function, selection_type, test_type')


# TODO: Use this component in OWDifferentialExpression

class GeneScoringWidget(QWidget):

    # default methods, override to add custom scoring methods
    scores = [gene_scoring_method('T-test', score_t_test, TwoTail, TwoSampleTest),
              gene_scoring_method('Mann-Whitney', score_mann_whitney, LowTail, TwoSampleTest),
              gene_scoring_method('Hypergeometric Test', score_hypergeometric_test, TwoTail, TwoSampleTest)]

    def __init__(self, box, parent,  **kwargs):
        # type: (Union[QGroupBox, QWidget], QWidget) -> None
        super().__init__(**kwargs)

        self.widget = box
        self.parent = parent

        # parent widget settings
        self.scoring_method_selection = None
        self.scoring_method_design = None
        self.test_type = None

    def set_method_selection_area(self, settings_var):
        # type: (str) -> None
        self.scoring_method_selection = settings_var

        comboBox(self.widget, self.parent, self.scoring_method_selection,
                 items=[sm.name for sm in self.scores],
                 callback=self.on_method_selection_changed,
                 label='Method')

    def set_method_design_area(self, settings_var):
        # type: (str) -> None
        self.scoring_method_design = settings_var

        radioButtons(self.widget, self.parent, self.scoring_method_design,
                     ['Cluster vs. rest', 'Cluster vs. cluster (max p-value)'],
                     callback=self.on_design_selection_changed,
                     label='Design')

    def set_test_type(self, settings_var):
        # type: (str) -> None
        self.test_type = settings_var
        radioButtons(self.widget, self.parent, self.test_type,
                     list(map(lambda s: s.title(), ALTERNATIVES)),
                     callback=self.on_test_type_changed,
                     label='Test Type')

    def get_selected_method(self):
        # type: () -> gene_scoring_method
        return self.scores[self.parent.__getattribute__(self.scoring_method_selection)]

    def get_selected_desig(self):
        # type: () -> int
        return self.parent.__getattribute__(self.scoring_method_design)

    def get_selected_test_type(self):
        # type: () -> str
        return ALTERNATIVES[self.parent.__getattribute__(self.test_type)]

    def on_method_selection_changed(self):
        """ Override this method if needed """
        self.parent.invalidate()

    def on_design_selection_changed(self):
        """ Override this method if needed """
        self.parent.invalidate()

    def on_test_type_changed(self):
        """ Override this method if needed """
        self.parent.invalidate()



import unittest
import tempfile
import shutil

try:
    from unittest import mock
except ImportError:
    import backports.unittest_mock
    backports.unittest_mock.install()
    from unittest import mock

try:
    from types import SimpleNamespace as namespace
except ImportError:
    class namespace(object):
        def __init__(self, **kwargs): self.__dict__.update(kwargs)
        def __repr__(self):
            contents = ",".join("{}={!r}".format(*it)
                                for it in sorted(self.__dict__.items()))
            return "namespace(" + contents + ")"

import doctest

from orangecontrib.bio import kegg
from orangecontrib.bio.kegg import api as keggapi
from orangecontrib.bio.kegg import conf as keggconf


list_organism = """\
T01001\thsa\tHomo sapiens (human)\tEukaryotes;Animals;Vertebrates;Mammals
T00005\tsce\tSaccharomyces cerevisiae (budding yeast)\tEukaryotes;Fungi;Ascomycetes;Saccharomycetes
T00245\tddi\tDictyostelium discoideum (cellular slime mold)\tEukaryotes;Protists;Amoebozoa;Dictyostelium\
"""

list_pathway_hsa = """\
path:hsa00010\tGlycolysis / Gluconeogenesis - Homo sapiens (human)
path:hsa00020\tCitrate cycle (TCA cycle) - Homo sapiens (human)
path:hsa00030\tPentose phosphate pathway - Homo sapiens (human)\
"""

info_pathway = """\
pathway          KEGG Pathway Database
path             Release 81.0+/01-18, Jan 17
                 Kanehisa Laboratories
                 479,620 entries
"""

genome_T01001 = """\
ENTRY       T01001            Complete  Genome
NAME        hsa, HUMAN, 9606
DEFINITION  Homo sapiens (human)
ANNOTATION  manual
TAXONOMY    TAX:9606
  LINEAGE   Eukaryota; Metazoa; Chordata; Craniata; Vertebrata; Euteleostomi; Mammalia; Eutheria; Euarchontoglires; Primates; Haplorrhini; Catarrhini; Hominidae; Homo
DATA_SOURCE RefSeq (Assembly:GCF_000001405.31)
ORIGINAL_DB NCBI
            OMIM
            HGNC
            HPRD
            Ensembl
STATISTICS  Number of protein genes:       20234
            Number of RNA genes:           18981
///
"""


def mock_service():
    s = namespace(
        list=namespace(
            organism=namespace(get=lambda: list_organism),
            pathway=lambda org: {
                "hsa": namespace(get=lambda: list_pathway_hsa)
            }[org],
        ),
        info=lambda db:
            {"pathway": namespace(get=lambda: info_pathway)}[db],
        get=lambda key: {
            "genome:T01001": namespace(get=lambda: genome_T01001)
        }[key],
    )
    return s


def mock_kegg_api():
    api = keggapi.KeggApi()
    api.service = mock_service()
    return api


def load_tests(loader, tests, ignore):
    def setUp(testcase):
        # testcase._tmpdir = tempfile.TemporaryDirectory(prefix="kegg-tests")
        testcase._tmpdir = tempfile.mkdtemp(prefix="kegg-tests")
        testcase._old_cache_path = keggconf.params["cache.path"]
        keggconf.params["cache.path"] = testcase._tmpdir
        testcase._mock_ctx = mock.patch(
            "orangecontrib.bio.kegg.api.web_service",
            mock_service)
        testcase._mock_ctx.__enter__()
        s = keggapi.web_service()
        assert isinstance(s, namespace)

    def tearDown(testcase):
        testcase._mock_ctx.__exit__(None, None, None)
        keggconf.params["cache.path"] = testcase._old_cache_path
        shutil.rmtree(testcase._tmpdir)

    api = mock_kegg_api()

    tests.addTests(
        doctest.DocTestSuite(
            keggapi, optionflags=doctest.ELLIPSIS,
            extraglobs={"api": api},
            setUp=setUp, tearDown=tearDown
        )
    )
    return tests



import unittest

from orangecontrib.bioinformatics.ncbi import taxonomy


class TestTaxonomy(unittest.TestCase):

    human = '9606'
    dicty = '352472'

    def test_common_taxonomy(self):
        self.assertGreater(len(taxonomy.common_taxids()), 0)

        self.assertEqual(taxonomy.name(self.human), 'Homo sapiens')
        self.assertEqual(taxonomy.name(self.dicty), 'Dictyostelium discoideum AX4')

        self.assertEqual(taxonomy.taxname_to_taxid('Homo sapiens'), self.human)
        self.assertEqual(taxonomy.taxname_to_taxid('Dictyostelium discoideum AX4'), self.dicty)

        self.assertGreater(len(taxonomy.shortname(self.human)), 0)
        self.assertGreater(len(taxonomy.shortname(self.dicty)), 0)

    def test_human(self):
        tax_obj = taxonomy.Taxonomy()
        self.assertGreater(len(tax_obj.taxids()), 1500000)

        self.assertTrue(isinstance(tax_obj.get_entry(self.human), taxonomy.utils.Taxon))
        self.assertRaises(taxonomy.utils.UnknownSpeciesIdentifier, tax_obj.get_entry, 'unknown_tax')

        self.assertTrue(('man', 'common name') in tax_obj.other_names(self.human))
        self.assertEqual(tax_obj.rank(self.human), 'species')
        self.assertEqual(tax_obj.parent(self.human), '9605')
        self.assertGreater(len(tax_obj.search('Homo sapiens', exact=True)), 0)
        self.assertGreater(len((tax_obj.lineage(self.human))), 0)

        self.assertGreater(len(tax_obj.get_all_strains(self.human)), 0)


if __name__ == '__main__':
    unittest.main()



import os

from Orange.misc.environ import data_dir


def progress_bar_milestones(count, iterations=100):
    return set([int(i*count/float(iterations)) for i in range(iterations)])


local_cache = os.path.join(data_dir(), 'bioinformatics/')


def ensure_type(value, types):
    if isinstance(value, types):
        return value
    else:
        raise TypeError('Wrong variable type. {value} is {value_type}, but should be {types}'.format(
            value=value, value_type=type(value), types=types))



import unittest

from AnyQt.QtCore import QStringListModel
from orangecontrib.bioinformatics.widgets.utils.gui import TokenListCompleter


class TestCompleter(unittest.TestCase):
    def test_TokenListCompleter(self):
        from Orange.widgets.utils.itemmodels import PyListModel
        completer = TokenListCompleter()
        completer.setTokenList(["foo", "bar", "baz"])
        completer.setCompletionPrefix("foo b")

        def completions(completer):
            current = completer.currentRow()
            items = []
            for i in range(completer.completionCount()):
                completer.setCurrentRow(i)
                items.append(completer.currentCompletion())
            completer.setCurrentRow(current)
            return items

        self.assertSequenceEqual(completions(completer),
                                 ["foo bar", "foo baz"])
        completer.setModel(None)
        self.assertSequenceEqual(completer.tokenList(), [])
        self.assertSequenceEqual(completions(completer), [])

        completer.setModel(QStringListModel(["a", "ab", "b"]))
        self.assertSequenceEqual(completer.tokenList(), ["a", "ab", "b"])
        completer.setCompletionPrefix("a a")
        self.assertSequenceEqual(completions(completer), ["a a", "a ab"])

        completer.setModel(PyListModel(["a", "aa", "ab", "ad"]))
        self.assertSequenceEqual(completer.tokenList(),
                                 ["a", "aa", "ab", "ad"])

        completer.model()[-1] = "z"
        self.assertSequenceEqual(completer.tokenList(),
                                 ["a", "aa", "ab", "z"])

        completer.model()[-2:] = ["ax", "az"]
        self.assertSequenceEqual(completer.tokenList(),
                                 ["a", "aa", "ax", "az"])

        completer.setSeparator(",")
        completer.setCompletionPrefix("a, a")
        self.assertSequenceEqual(completions(completer),
                                 ["a, a", "a, aa", "a, ax", "a, az"])
